import sys

if sys.version_info >= (3, 8):
  from importlib.metadata import PackageNotFoundError
  from importlib.metadata import version as importlib_version
else:
  from importlib_metadata import PackageNotFoundError
  from importlib_metadata import version as importlib_version

try:
  _build_version = importlib_version('comfy_catapult')
except PackageNotFoundError:
  _build_version = '0.0.0'



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import json
from typing import List, Literal
from urllib.parse import ParseResult, urljoin, urlparse, urlunparse

from ..errors import (BasedURLValidationError, URLDirectoryValidationError,
                      URLValidationError)

ComfyAPIScheme = Literal['http', 'https']
VALID_COMFY_API_SCHEMES: List[ComfyAPIScheme] = ['http', 'https']


def SmartURLJoin(base: str, url: str) -> str:
  """urljoin() but can handle relative paths even for custom schemes.

  See: https://github.com/python/cpython/issues/63028

  From: https://github.com/python/cpython/issues/63028#issuecomment-1564858715
  """
  base_pr = urlparse(base)
  bscheme = base_pr.scheme

  url_pr = urlparse(url)
  scheme = url_pr.scheme or bscheme
  if bscheme != scheme:
    return url

  base_pr = base_pr._replace(scheme='http')
  url_pr = url_pr._replace(scheme='http')

  joined = urljoin(urlunparse(base_pr), urlunparse(url_pr))
  joined_pr = urlparse(joined)
  joined_pr = joined_pr._replace(scheme=scheme)
  return urlunparse(joined_pr)


def JoinToBaseURL(base: str, path: str) -> str:
  """Takes a URL and appends the path to it.
  
  Always assumes the base is a directory. Always assumes the path is relative to that directory."""
  if not base.endswith('/'):
    base += '/'
  if path.startswith('/'):
    path = path[1:]
  return base + path


def IsValidURL(url: str) -> bool:
  try:
    urlparse(url)
    return True
  except ValueError:
    return False


def ToParseResult(url: str) -> ParseResult:
  try:
    return urlparse(url)
  except ValueError as e:
    raise URLValidationError(f'URL {json.dumps(url)} is not valid: {e}') from e


def ValidateIsURL(url: str) -> str:
  if not IsValidURL(url=url):
    raise URLValidationError(f'URL {json.dumps(url)} is not valid')
  return url


def IsWeaklyRelativeTo(*, base: str, url: str) -> bool:
  url = ValidateIsURL(url=url)
  base = ValidateIsURL(url=base)
  base_parsed = ToParseResult(url=base)
  joined = SmartURLJoin(base, url)
  joined_parsed = ToParseResult(url=joined)

  if (base_parsed.scheme, base_parsed.netloc) != (joined_parsed.scheme,
                                                  joined_parsed.netloc):
    return False

  return joined_parsed.path.startswith(base_parsed.path)


def ValidateIsBasedURL(*, url: str, any_bases: List[str]) -> str:
  url = ValidateIsURL(url=url)

  for base in any_bases:
    base = ValidateIsURL(url=base)
    if IsWeaklyRelativeTo(base=base, url=url):
      return url
  raise BasedURLValidationError(
      f'URL {json.dumps(url)} is not relative to any of {any_bases}')


def Relativize(*, base: str, url: str) -> str:
  """Return the relative path from base to url, as a valid relative URL.

  """
  url = ValidateIsURL(url=url)
  base = ValidateIsURL(url=base)
  url = ValidateIsBasedURL(url=url, any_bases=[base])

  url_parsed = ToParseResult(url=url)
  base_parsed = ToParseResult(url=base)

  url_path = url_parsed.path
  base_path = base_parsed.path
  if not url_path.startswith(base_path):
    raise ValueError(
        f'URL {json.dumps(url)} is not relative to base {json.dumps(base)}')

  return url_path[len(base_path):]


def ValidateIsURLDirectory(url: str) -> str:
  url = ValidateIsURL(url=url)
  url_pr: ParseResult = ToParseResult(url=url)
  if not url_pr.path.endswith('/'):
    raise URLDirectoryValidationError(
        f'URL {json.dumps(url)} is not a directory, because it does not end with a trailing slash'
    )
  return url


def ValidateIsComfyAPITargetURL(url: str) -> str:
  url_pr: ParseResult = ToParseResult(url=url)
  if url_pr.scheme not in VALID_COMFY_API_SCHEMES:
    raise ValueError(
        f'URL {json.dumps(url)} is not a comfy API target URL, because'
        f' its scheme is not one of {VALID_COMFY_API_SCHEMES}')
  if url_pr.hostname is None or url_pr.hostname == '':
    raise ValueError(
        f'URL {json.dumps(url)} is not a comfy API target URL, because'
        f' its hostname is empty')

  return url



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import base64
import json
import logging
import textwrap
from typing import Any, Dict, List, Optional, Type, TypeVar, Union
from urllib.parse import urlencode, urlparse

import aiohttp
from anyio import Path
from pydantic import BaseModel

from ._internal.url_utils import JoinToBaseURL
from ._internal.utilities import DumpYaml, TryParseAsModel, WatchVar
from .api_client_base import ComfyAPIClientBase
from .comfy_schema import (APIHistory, APIObjectInfo, APIPromptInfo,
                           APIQueueInfo, APISystemStats, APIUploadImageResp,
                           APIWorkflowTicket, ClientID, PromptID)

logger = logging.getLogger(__name__)

T = TypeVar('T')


async def _TryGetContent(*,
                         resp: aiohttp.ClientResponse) -> Union[str, Exception]:
  try:
    return await resp.text(errors='replace')
  except Exception as e:
    logger.exception('Error getting content', exc_info=e, stack_info=True)
    return e


async def _RaiseForStatus(*,
                          resp: aiohttp.ClientResponse,
                          extra: Optional[Any] = None):
  """Same as raise_for_status(), but also dumps any content in the response into the exception"""
  try:
    resp.raise_for_status()
  except aiohttp.ClientResponseError as e:
    content_or_exc = await _TryGetContent(resp=resp)

    message = e.message
    if isinstance(content_or_exc, str):
      message += f'\n\nContent (raw): {textwrap.indent(content_or_exc, prefix="  ")}'
    else:
      message += f'\n\nError getting content: {content_or_exc}'
    if extra is not None:
      message += f'\n\nExtra info: {textwrap.indent(await DumpYaml(extra), prefix="  ")}'

    raise aiohttp.ClientResponseError(request_info=e.request_info,
                                      history=e.history,
                                      code=e.code,
                                      status=e.status,
                                      message=message,
                                      headers=e.headers) from e


async def _TryParseAsJson(*, content: str, json_type: Type[T]) -> T:
  try:
    result = json.loads(content)
    if not isinstance(result, json_type):
      raise TypeError(f'Expected {json_type}, got {type(result)}')
    return result
  except json.JSONDecodeError as e:
    raise Exception(
        f'Error: {e}\n\nContent (raw): {textwrap.indent(content, prefix="  ")}'
    ) from e


async def _TryParseRespAsJson(*, resp: aiohttp.ClientResponse,
                              json_type: Type[T]) -> T:
  content_bytes = b''
  content_str: str = ''
  content: Optional[T] = None
  try:
    content_bytes = await resp.content.read()
    content_str = content_bytes.decode('utf-8')
    if resp.content_type != 'application/json':
      raise Exception(
          f'Error: {resp.status} {resp.reason}'
          f'\n\nExpected content-type: application/json, got {resp.content_type}'
          f'\n\nContent (raw):\n{textwrap.indent(content_str, prefix="  ")}')

    content = await _TryParseAsJson(content=content_str, json_type=json_type)
    if resp.ok:
      await _RaiseForStatus(resp=resp, extra=content)
    return content
  except Exception as e:
    contentb64 = base64.b64encode(content_bytes).decode('utf-8')
    raise Exception(
        f'Error: {resp.status} {resp.reason}'
        f'\n\nContent (raw):\n{textwrap.indent(content_str, prefix="  ")}'
        f'\n\nContent (yaml):\n{textwrap.indent(await DumpYaml(content), prefix="  ")}'
        f'\n\nContent (base64):\n{textwrap.indent(contentb64, prefix="  ")}'
    ) from e


_BaseModelT = TypeVar('_BaseModelT', bound=BaseModel)


async def _TryParseRespAsModel(
    *, resp: aiohttp.ClientResponse, model_type: Type[_BaseModelT],
    errors_dump_directory: Optional[Path]) -> _BaseModelT:
  content: Any = await _TryParseRespAsJson(resp=resp, json_type=dict)
  return await TryParseAsModel(content=content,
                               model_type=model_type,
                               errors_dump_directory=errors_dump_directory)


class ComfyAPIClient(ComfyAPIClientBase):

  def __init__(self,
               comfy_api_url: str,
               *,
               errors_dump_directory: Optional[Path] = None):
    self._comfy_api_url = comfy_api_url
    self._session = aiohttp.ClientSession()
    self._errors_dump_directory = errors_dump_directory

  async def __aenter__(self):
    await self._session.__aenter__()
    return self

  async def __aexit__(self, exc_type, exc, tb):
    await self._session.__aexit__(exc_type, exc, tb)
    await self.Close()

  async def Close(self):
    await self._session.close()

  def GetURL(self) -> str:
    return self._comfy_api_url

  async def GetSystemStatsRaw(self) -> dict:
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'system_stats'))
    async with WatchVar(url=url.geturl()):
      async with self._session.get(url.geturl()) as resp:
        return await _TryParseRespAsJson(resp=resp, json_type=dict)

  async def GetSystemStats(self) -> APISystemStats:
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'system_stats'))
    async with WatchVar(url=url.geturl()):
      async with self._session.get(url.geturl()) as resp:
        return await _TryParseRespAsModel(
            resp=resp,
            model_type=APISystemStats,
            errors_dump_directory=self._errors_dump_directory)

  async def GetObjectInfoRaw(self) -> dict:
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'object_info'))
    async with WatchVar(url=url.geturl()):
      async with self._session.get(url.geturl()) as resp:
        return await _TryParseRespAsJson(resp=resp, json_type=dict)

  async def GetObjectInfo(self) -> APIObjectInfo:
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'object_info'))
    async with WatchVar(url=url.geturl()):
      async with self._session.get(url.geturl()) as resp:
        return await _TryParseRespAsModel(
            resp=resp,
            model_type=APIObjectInfo,
            errors_dump_directory=self._errors_dump_directory)

  async def GetPromptRaw(self) -> dict:
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'prompt'))
    async with WatchVar(url=url.geturl()):
      async with self._session.get(url.geturl()) as resp:
        return await _TryParseRespAsJson(resp=resp, json_type=dict)

  async def GetPrompt(self) -> APIPromptInfo:
    prompt_info_dict = await self.GetPromptRaw()
    return await TryParseAsModel(
        content=prompt_info_dict,
        model_type=APIPromptInfo,
        errors_dump_directory=self._errors_dump_directory)

  async def PostPromptRaw(self,
                          *,
                          prompt_workflow: dict,
                          number: Optional[int] = None,
                          client_id: Optional[ClientID] = None,
                          prompt_id: Optional[PromptID] = None,
                          extra_data: Optional[dict] = None) -> dict:
    body: Dict[str, Any] = {'prompt': prompt_workflow}

    if number is not None:
      body['number'] = number
    if client_id is not None:
      body['client_id'] = client_id
    if prompt_id is not None:
      body['prompt_id'] = prompt_id
    if extra_data is not None:
      body['extra_data'] = extra_data

    data: bytes = json.dumps(body).encode('utf-8')
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'prompt'))
    async with WatchVar(url=url.geturl()):
      async with self._session.post(url.geturl(), data=data) as resp:
        return await _TryParseRespAsJson(resp=resp, json_type=dict)

  async def PostPrompt(self,
                       *,
                       prompt_workflow: dict,
                       number: Optional[int] = None,
                       client_id: Optional[ClientID] = None,
                       prompt_id: Optional[PromptID] = None,
                       extra_data: Optional[dict] = None) -> APIWorkflowTicket:
    ticket = await self.PostPromptRaw(prompt_workflow=prompt_workflow,
                                      number=number,
                                      client_id=client_id,
                                      prompt_id=prompt_id,
                                      extra_data=extra_data)
    return await TryParseAsModel(
        content=ticket,
        model_type=APIWorkflowTicket,
        errors_dump_directory=self._errors_dump_directory)

  async def GetHistoryRaw(self,
                          *,
                          prompt_id: Optional[PromptID] = None,
                          max_items: Optional[int] = None) -> dict:
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'history'))
    if max_items is not None:
      url = url._replace(query=f'max_items={max_items}')
    if prompt_id is not None:
      url = url._replace(path=f'{url.path}/{prompt_id}')

    async with WatchVar(url=url.geturl()):
      async with self._session.get(url.geturl()) as resp:
        return await _TryParseRespAsJson(resp=resp, json_type=dict)

  async def GetHistory(self,
                       *,
                       prompt_id: Optional[PromptID] = None,
                       max_items: Optional[int] = None) -> APIHistory:
    history = await self.GetHistoryRaw(prompt_id=prompt_id, max_items=max_items)
    return await TryParseAsModel(
        content=history,
        model_type=APIHistory,
        errors_dump_directory=self._errors_dump_directory)

  async def GetQueueRaw(self) -> dict:
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'queue'))
    async with WatchVar(url=url.geturl()):
      async with self._session.get(url.geturl()) as resp:
        return await _TryParseRespAsJson(resp=resp, json_type=dict)

  async def GetQueue(self) -> APIQueueInfo:
    queue: dict = await self.GetQueueRaw()
    return await TryParseAsModel(
        content=queue,
        model_type=APIQueueInfo,
        errors_dump_directory=self._errors_dump_directory)

  async def PostUploadImageRaw(self, *, folder_type: str, subfolder: str,
                               filename: str, data: bytes,
                               overwrite: bool) -> dict:
    """

    * See https://github.com/comfyanonymous/ComfyUI/blob/0c2c9fbdfa53c2ad3b7658a7f2300da831830388/server.py#L203

    * See https://github.com/comfyanonymous/ComfyUI/blob/0c2c9fbdfa53c2ad3b7658a7f2300da831830388/server.py#L161

    Args:
        folder_type (str): _description_
        subfolder (str): _description_
        filename (str): _description_
        data (bytes): _description_
        overwrite (bool): _description_

    Returns:
        dict: _description_
    """
    async with WatchVar(folder_type=folder_type,
                        subfolder=subfolder,
                        filename=filename,
                        overwrite=overwrite):
      fdata = aiohttp.FormData()
      fdata.add_field('image',
                      data,
                      filename=filename,
                      content_type='application/octet-stream')
      fdata.add_field('overwrite', str(overwrite).lower())
      fdata.add_field('subfolder', subfolder)
      fdata.add_field('type', folder_type)
      fdata.add_field('filename', filename)
      post_url = urlparse(JoinToBaseURL(self._comfy_api_url, 'upload/image'))
      async with WatchVar(post_url=post_url.geturl(), fdata=fdata):
        async with self._session.post(post_url.geturl(), data=fdata) as resp:
          result = await _TryParseRespAsJson(resp=resp, json_type=dict)
          if not isinstance(result, dict):
            # Server should never return a list or something other than a
            # dictionary.
            raise TypeError(f'Expected dict, got {type(result)}')
          return result

  async def PostUploadImage(self, *, folder_type: str, subfolder: str,
                            filename: str, data: bytes,
                            overwrite: bool) -> APIUploadImageResp:
    result = await self.PostUploadImageRaw(folder_type=folder_type,
                                           subfolder=subfolder,
                                           filename=filename,
                                           data=data,
                                           overwrite=overwrite)
    return await TryParseAsModel(
        content=result,
        model_type=APIUploadImageResp,
        errors_dump_directory=self._errors_dump_directory)

  async def GetView(self, *, folder_type: str, subfolder: str,
                    filename: str) -> bytes:
    data = {'filename': filename, 'subfolder': subfolder, 'type': folder_type}
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'view'))
    url = url._replace(query=urlencode(data))

    async with WatchVar(url=url.geturl()):
      async with self._session.get(url.geturl()) as resp:
        await _RaiseForStatus(resp=resp)
        return await resp.content.read()

  async def PostFree(self, *, unload_models: bool, free_memory: bool):
    data = {'unload_models': unload_models, 'free_memory': free_memory}
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'free'))
    async with WatchVar(url=url.geturl()):
      async with self._session.post(url.geturl(), data=data) as resp:
        await _RaiseForStatus(resp=resp)

  async def PostInterrupt(self):
    # TODO(realazthat/comfy-catapult#5): change the API to take a prompt_id.
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'interrupt'))
    async with WatchVar(url=url.geturl()):
      async with self._session.post(url.geturl()) as resp:
        resp.raise_for_status()

  async def PostQueue(self, *, delete: List[PromptID], clear: bool):
    url = urlparse(JoinToBaseURL(self._comfy_api_url, 'queue'))
    data = {'delete': delete, 'clear': clear}
    async with WatchVar(url=url.geturl()):
      async with self._session.post(url.geturl(), data=data) as resp:
        resp.raise_for_status()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import os
import unittest
from tempfile import TemporaryDirectory
from typing import List
from unittest import IsolatedAsyncioTestCase

import pydantic
from anyio import Path

from ._internal.url_utils import SmartURLJoin
from .comfy_schema import (VALID_FOLDER_TYPES, ComfyFolderType,
                           ComfyUIPathTriplet)
from .comfy_schema_test import VALID_SUBFOLDER_EDGES
from .remote_file_api_comfy import (ComfySchemeRemoteFileAPI,
                                    TripletToComfySchemeURL)

COMFY_API_URL = os.environ.get('COMFY_API_URL')
if COMFY_API_URL is None:
  raise ValueError('Please set COMFY_API_URL in the environment')


class TestRemoteFileApiComfy(IsolatedAsyncioTestCase):

  async def asyncSetUp(self):
    if COMFY_API_URL is None:
      raise ValueError('Please set COMFY_API_URL in the environment')
    self._comfy_api_url: str = COMFY_API_URL
    self._remote = ComfySchemeRemoteFileAPI(comfy_api_urls=[COMFY_API_URL],
                                            overwrite=True)

  async def asyncTearDown(self):
    pass

  async def _test_UploadFile(self, *, comfy_api_url: str,
                             triplet: ComfyUIPathTriplet):
    with TemporaryDirectory() as tmp_dir:
      path = Path(tmp_dir) / 'local-file.txt'
      local_download_path = Path(tmp_dir) / 'local-downloaded.txt'
      contents = 'hello world'
      await path.write_text(contents)

      input_url = TripletToComfySchemeURL(comfy_api_url=comfy_api_url,
                                          triplet=triplet)

      uploaded_url = await self._remote.UploadFile(src_path=path,
                                                   untrusted_dst_url=input_url)

      await self._remote.DownloadFile(untrusted_src_url=uploaded_url,
                                      dst_path=local_download_path)

      downloaded_contents = await local_download_path.read_text()
      self.assertEqual(contents, downloaded_contents)

  async def test_UploadFile(self):
    folder_type: ComfyFolderType
    folder_types: List[ComfyFolderType] = ['input']
    for folder_type in folder_types:
      for subfolder, _ in VALID_SUBFOLDER_EDGES:
        with self.subTest(folder_type=folder_type, subfolder=subfolder):
          await self._test_UploadFile(comfy_api_url=self._comfy_api_url,
                                      triplet=ComfyUIPathTriplet(
                                          type=folder_type,
                                          subfolder=subfolder,
                                          filename='remote-file.txt'))

  async def _test_UploadToTriplet(self, *, comfy_api_url: str,
                                  triplet: ComfyUIPathTriplet):
    with TemporaryDirectory() as tmp_dir:
      path = Path(tmp_dir) / 'local-file.txt'
      local_download_path = Path(tmp_dir) / 'local-downloaded.txt'
      contents = 'hello world'
      await path.write_text(contents)

      uploaded_triplet = await self._remote.UploadToTriplet(
          src_path=path,
          untrusted_comfy_api_url=comfy_api_url,
          untrusted_dst_triplet=triplet)

      await self._remote.DownloadTriplet(untrusted_comfy_api_url=comfy_api_url,
                                         untrusted_src_triplet=uploaded_triplet,
                                         dst_path=local_download_path)

      downloaded_contents = await local_download_path.read_text()
      self.assertEqual(contents, downloaded_contents)

  async def test_UploadToTriplet(self):
    folder_type: ComfyFolderType
    folder_types: List[ComfyFolderType] = ['input']
    for folder_type in folder_types:
      for subfolder, _ in VALID_SUBFOLDER_EDGES:
        with self.subTest(folder_type=folder_type, subfolder=subfolder):

          await self._test_UploadToTriplet(comfy_api_url=self._comfy_api_url,
                                           triplet=ComfyUIPathTriplet(
                                               type=folder_type,
                                               subfolder=subfolder,
                                               filename='remote-file.txt'))

  async def test__TripletToComfySchemeURL(self):
    comfy_api_url = 'http://comfy_host:23534/'
    comfy_scheme_url = 'comfy+http://comfy_host:23534/'

    self.assertEqual(
        TripletToComfySchemeURL(comfy_api_url=comfy_api_url,
                                triplet=ComfyUIPathTriplet(
                                    type='input',
                                    subfolder='',
                                    filename='remote-file.txt')),
        SmartURLJoin(comfy_scheme_url, 'input/remote-file.txt'))
    with self.assertRaises(pydantic.ValidationError):
      _ = ComfyUIPathTriplet(type='input',
                             subfolder='/',
                             filename='remote-file.txt')

    self.assertEqual(
        TripletToComfySchemeURL(comfy_api_url=comfy_api_url,
                                triplet=ComfyUIPathTriplet(
                                    type='input',
                                    subfolder='subfolder',
                                    filename='remote-file.txt')),
        SmartURLJoin(comfy_scheme_url, 'input/subfolder/remote-file.txt'))
    with self.assertRaises(pydantic.ValidationError):
      _ = ComfyUIPathTriplet(type='input',
                             subfolder='/subfolder',
                             filename='remote-file.txt')
    with self.assertRaises(pydantic.ValidationError):
      _ = ComfyUIPathTriplet(type='input',
                             subfolder='/subfolder/',
                             filename='remote-file.txt')
    self.assertEqual(
        TripletToComfySchemeURL(comfy_api_url=comfy_api_url,
                                triplet=ComfyUIPathTriplet(
                                    type='input',
                                    subfolder='subfolder/subsubfolder',
                                    filename='remote-file.txt')),
        SmartURLJoin(comfy_scheme_url,
                     'input/subfolder/subsubfolder/remote-file.txt'))

  async def test_TripletToComfySchemeURL(self):
    for comfy_api_url in [
        'http://comfy_host:23534',
    ]:
      for folder_type in VALID_FOLDER_TYPES:
        with self.subTest(folder_type=folder_type):
          self.assertEqual(
              TripletToComfySchemeURL(comfy_api_url=comfy_api_url,
                                      triplet=ComfyUIPathTriplet(
                                          type=folder_type,
                                          subfolder='',
                                          filename='remote-file.txt')),
              f'comfy+{comfy_api_url}/{folder_type}/remote-file.txt')
          self.assertEqual(
              TripletToComfySchemeURL(comfy_api_url=comfy_api_url,
                                      triplet=ComfyUIPathTriplet(
                                          type=folder_type,
                                          subfolder='subfolder',
                                          filename='remote-file.txt')),
              f'comfy+{comfy_api_url}/{folder_type}/subfolder/remote-file.txt')
          self.assertEqual(
              TripletToComfySchemeURL(comfy_api_url=comfy_api_url,
                                      triplet=ComfyUIPathTriplet(
                                          type=folder_type,
                                          subfolder='subfolder/subsubfolder',
                                          filename='remote-file.txt')),
              f'comfy+{comfy_api_url}/{folder_type}/subfolder/subsubfolder/remote-file.txt'
          )
          self.assertEqual(
              TripletToComfySchemeURL(comfy_api_url=comfy_api_url,
                                      triplet=ComfyUIPathTriplet(
                                          type=folder_type,
                                          subfolder='subfolder/subsubfolder/',
                                          filename='remote-file.txt')),
              f'comfy+{comfy_api_url}/{folder_type}/subfolder/subsubfolder/remote-file.txt'
          )


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import json
from copy import deepcopy
from typing import Optional, Sequence, Union

from .comfy_schema import APINodeID, APIWorkflowTicket


class CatapultRuntimeError(RuntimeError):
  pass


class NodeNotFound(CatapultRuntimeError):

  def __init__(self, *, node_id: Union[APINodeID, int, None],
               title: Union[str, int, None]):
    super().__init__(
        f'Node with title=={json.dumps(title)}, node_id=={json.dumps(node_id)} not found'
    )
    self.node_id = node_id
    self.title = title


class MultipleNodesFound(CatapultRuntimeError):

  def __init__(self, *, search_titles: Sequence[str],
               search_nodes: Sequence[APINodeID], found_titles: Sequence[str],
               found_nodes: Sequence[APINodeID]):
    super().__init__(
        f'Found multiple nodes with titles=={repr(found_titles)}, node_ids=={list(found_nodes)}'
        f' when searching for titles=={repr(search_titles)}, node_ids=={list(search_nodes)}'
    )
    self.search_titles = list(search_titles)
    self.search_node_ids = list(search_nodes)
    self.found_titles = list(found_titles)
    self.found_node_ids = list(found_nodes)


class NodesNotExecuted(CatapultRuntimeError):

  def __init__(self, *, nodes: Sequence[APINodeID],
               titles: Optional[Sequence[Optional[str]]]):
    if titles is None:
      titles = [None] * len(nodes)
    if len(nodes) != len(titles):
      raise ValueError(
          f'len(nodes) != len(titles): {len(nodes)} != {len(titles)}')

    super().__init__(
        f'Nodes {",".join(map(repr, nodes))} with titles {",".join(map(repr,titles))} not executed'
    )
    self.nodes = list(nodes)
    self.titles = list(titles)


class WorkflowSubmissionError(CatapultRuntimeError):

  def __init__(self, msg, *, prepared_workflow: dict,
               ticket: APIWorkflowTicket):
    super().__init__(msg)
    self.prepared_workflow: dict = deepcopy(prepared_workflow)
    self.ticket: APIWorkflowTicket = deepcopy(ticket)


class JobNotFound(CatapultRuntimeError):
  pass


class JobFailed(CatapultRuntimeError):
  pass


class URLValidationError(ValueError):
  pass


class BasedURLValidationError(ValueError):
  pass


class URLDirectoryValidationError(ValueError):
  pass



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

from abc import ABC, abstractmethod
from typing import List, Tuple

from anyio import Path

from .comfy_schema import ComfyUIPathTriplet


class RemoteFileAPIBase(ABC):

  @abstractmethod
  async def UploadFile(self, *, src_path: Path, untrusted_dst_url: str) -> str:
    """Upload a file.

    Args:
        src_path (Path): _description_
        untrusted_dst_url (str): _description_


    Returns:
        str: A new URL, sometimes the stored URL is not the same as the uploaded
        one, e.g if there is already an existing file there.
    """
    raise NotImplementedError()

  async def UploadToTriplet(
      self, *, src_path: Path, untrusted_comfy_api_url: str,
      untrusted_dst_triplet: ComfyUIPathTriplet) -> ComfyUIPathTriplet:
    """Upload a file to a (comfy api server, folder_type, subfolder, filename).

    Raises:
        NotImplementedError: _description_

    Returns:
        ComfyUIPathTriplet: A new triplet, sometimes the stored triplet is not
        the same as the uploaded one, e.g if there is already an existing file
        there.
    """
    raise NotImplementedError()

  @abstractmethod
  async def DownloadFile(self, *, untrusted_src_url: str, dst_path: Path):
    raise NotImplementedError()

  @abstractmethod
  async def DownloadTriplet(self, *, untrusted_comfy_api_url: str,
                            untrusted_src_triplet: ComfyUIPathTriplet,
                            dst_path: Path):
    raise NotImplementedError()

  @abstractmethod
  def TripletToURL(self, *, comfy_api_url: str,
                   triplet: ComfyUIPathTriplet) -> str:
    """Convert a triplet, that this API can handle, to a URL that this API can handle."""
    raise NotImplementedError()

  @abstractmethod
  def URLToTriplet(self, *, url: str) -> Tuple[str, ComfyUIPathTriplet]:
    """Convert a URL that this API can handle to a triplet, that this API can handle."""
    raise NotImplementedError()

  @abstractmethod
  def GetBases(self) -> List[str]:
    """Return a list of base URLs that this API can handle."""
    raise NotImplementedError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import logging
from typing import (Generator, Hashable, List, Literal, NamedTuple, Optional,
                    Union, cast)

import pydash
from anyio import Path

from .comfy_schema import (APIHistoryEntry, APINodeID, APIOutputUI,
                           APIWorkflow, APIWorkflowNodeInfo,
                           ComfyUIPathTriplet)
from .errors import MultipleNodesFound, NodeNotFound
from .remote_file_api_base import RemoteFileAPIBase

logger = logging.getLogger(__name__)


class NodeIDAndNode(NamedTuple):
  node_id: APINodeID
  node_info: APIWorkflowNodeInfo


def FindNodesByTitle(*, workflow: APIWorkflow,
                     title: str) -> Generator[NodeIDAndNode, None, None]:
  node_id: APINodeID
  node_info: APIWorkflowNodeInfo
  for node_id, node_info in workflow.root.items():
    if node_info.meta is not None and node_info.meta.title == title:
      yield NodeIDAndNode(node_id=node_id, node_info=node_info)


def FindNodeByTitle(*, workflow: APIWorkflow,
                    title: str) -> Optional[NodeIDAndNode]:
  for (node_id, node_info) in FindNodesByTitle(workflow=workflow, title=title):
    return NodeIDAndNode(node_id=node_id, node_info=node_info)
  return None


def GetNodeByTitle(*, workflow: APIWorkflow, title: str) -> NodeIDAndNode:
  nodes: List[NodeIDAndNode] = list(
      FindNodesByTitle(workflow=workflow, title=title))

  if len(nodes) == 0:
    raise NodeNotFound(title=title, node_id=None)

  if len(nodes) > 1:
    raise MultipleNodesFound(search_titles=[title],
                             search_nodes=[title],
                             found_titles=[],
                             found_nodes=[node_id for node_id, _ in nodes])
  node_id, node_info = nodes[0]

  return NodeIDAndNode(node_id=node_id, node_info=node_info)


def FindNode(*, workflow: APIWorkflow,
             id_or_title: Union[int, str]) -> Optional[NodeIDAndNode]:
  id_node_id: Optional[APINodeID] = None
  # id_node_error: Optional[Exception] = None

  title_node_id: Optional[APINodeID] = None
  # title_node_error: Optional[Exception] = None

  try:
    if isinstance(id_or_title, str):
      title_node_id, _ = GetNodeByTitle(workflow=workflow, title=id_or_title)

  except NodeNotFound:
    # id_node_error = e
    pass

  id_or_title_str = str(id_or_title)
  try:
    id_node_id_: APINodeID = cast(APINodeID, id_or_title_str)
    if id_node_id_ in workflow.root:
      id_node_id = id_node_id_
  except ValueError:
    # title_node_error = e
    pass

  if title_node_id is None and id_node_id is None:
    return None
  elif title_node_id is not None and id_node_id is not None:
    raise MultipleNodesFound(search_titles=[id_or_title_str],
                             search_nodes=[id_node_id],
                             found_titles=[title_node_id],
                             found_nodes=[id_node_id])
  elif title_node_id is not None:
    return NodeIDAndNode(node_id=title_node_id,
                         node_info=workflow.root[title_node_id])
  else:
    if id_node_id is None:
      raise AssertionError('id_node_id is None')

    return NodeIDAndNode(node_id=id_node_id,
                         node_info=workflow.root[id_node_id])


def GetNode(*, workflow: APIWorkflow, id_or_title: Union[str,
                                                         int]) -> NodeIDAndNode:
  node = FindNode(workflow=workflow, id_or_title=id_or_title)
  if node is None:
    raise NodeNotFound(title=id_or_title, node_id=id_or_title)
  return node


def GenerateNewNodeID(*, workflow: APIWorkflow) -> APINodeID:
  all_keys = workflow.root.keys()
  max_integer = 0

  # Filter all keys that fail to convert to int
  for key in all_keys:
    try:
      max_integer = max(int(key), max_integer)
    except ValueError:
      continue

  return cast(APINodeID, str(max_integer + 1))


async def DownloadPreviewImage(*, node_id: APINodeID,
                               job_history: APIHistoryEntry,
                               field_path: Union[Hashable, List[Hashable]],
                               comfy_api_url: str, remote: RemoteFileAPIBase,
                               local_dst_path: Path):
  """Downloads something that looks like an Preview Image node's outputs.

  * field_path=='gifs[0]' works for Video Combine
  * field_path=='images[0]' works for Preview Image



  Example Preview Image Workflow API json:

  ```
  '25':
    inputs:
      images:
      - '8'
      - 0
    class_type: PreviewImage
    meta:
      title: Preview Image
  ```

  Example Preview Image node output:

  ```
  outputs:
    '25':
      images:
      - filename: ComfyUI_temp_huntb_00001_.png
        subfolder: ''
        type: temp
  ```

  Example Video Combine node output:

  TODO: Put an example here.

  Args:
      node_id: The node_id.
      job_history: The job_history.
      field_path: A pydash field path, for the pydash.get() and pydash.set_()
        functions.
      comfy_api_url: e.g http://127.0.0.1:8188.
      remote: A RemoteFileAPI instance.
      local_dst_path: Path to the local destination file.
  """

  if job_history.outputs is None:
    raise AssertionError('job_history.outputs is None')

  if node_id not in job_history.outputs:
    raise Exception(f'{node_id} not in job_history.outputs')

  node_outputs: APIOutputUI = job_history.outputs[node_id]

  file_dict: dict = pydash.get(node_outputs.root, field_path)

  if 'filename' not in file_dict:
    raise Exception(f'Expected "filename" in {file_dict}')
  filename: str = file_dict['filename']
  if not isinstance(filename, str):
    raise Exception(f'Expected "filename" to be str, got {type(filename)}')
  if 'subfolder' not in file_dict:
    raise Exception(f'Expected "subfolder" in {file_dict}')
  subfolder: str = file_dict['subfolder']
  if not isinstance(subfolder, str):
    raise Exception(f'Expected "subfolder" to be str, got {type(subfolder)}')
  if 'type' not in file_dict:
    raise Exception(f'Expected "type" in {file_dict}')
  folder_type: Literal['temp', 'output'] = file_dict['type']
  if not isinstance(folder_type, str):
    raise Exception(f'Expected "type" to be str, got {type(folder_type)}')
  if folder_type not in ['temp', 'output']:
    raise Exception(
        f'Expected "type" to be "temp" or "output", got {folder_type}')

  triplet = ComfyUIPathTriplet(type=folder_type,
                               subfolder=subfolder,
                               filename=filename)
  return await remote.DownloadTriplet(untrusted_comfy_api_url=comfy_api_url,
                                      untrusted_src_triplet=triplet,
                                      dst_path=local_dst_path)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import json
from typing import Any, Dict, List, Literal, NamedTuple, Optional, Union
from urllib.parse import urljoin

from pydantic import BaseModel, ConfigDict, Field, RootModel, field_validator
from typing_extensions import Annotated

EXTRA: Union[Literal['allow', 'ignore', 'forbid'], None] = 'allow'

APINodeID = Annotated[
    str,
    Field(alias='node_id', description='The ID of a node in a workflow.')]
PromptID = Annotated[
    str,
    Field(
        alias='prompt_id',
        description=
        'The ID of a prompt (submission of a api workflow to the server to be executed). You can choose this yourself when submitting.'
    )]
ClientID = Annotated[str, Field(alias='client_id')]
OutputName = Annotated[
    str,
    Field(
        alias='output_name',
        description=
        'The name of an output in a node, in /history/, /history/{prompt_id} endpoints.'
    )]
# This is BOOLEAN, INT etc.
OutputType = Annotated[
    str,
    Field(
        alias='output_type',
        description=
        'The type of a named output of a node, in /object_info endpoint, and also can be seen/found in {node,custom node} implementations.'
    )]

# This is BOOLEAN, INT etc.
NamedInputType = Annotated[
    str,
    Field(
        alias='input_type',
        description=
        'The type of a named input of a node, in /object_info endpoint, and also can be seen/found in {node,custom node} implementations.'
    )]
# This is a list of valid *values* for a combo input.
ComboInputType = Annotated[List[Any], Field(alias='combo_input_class')]
ComfyFolderType = Literal['input', 'output', 'temp']
VALID_FOLDER_TYPES: List[ComfyFolderType] = ['input', 'output', 'temp']


################################################################################
class APIWorkflowInConnection(NamedTuple):
  """Represents a connection between two nodes in a workflow. This is used in the input of a node."""
  output_node_id: APINodeID
  output_index: int


class APIWorkflowNodeMeta(BaseModel):
  """Nodes are allowed to have a `_meta` field.

  The meta field was was added in
  https://github.com/comfyanonymous/ComfyUI/pull/2380 for information such as
  the title of the node.
  """
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """
  title: Optional[str] = None


class APIWorkflowNodeInfo(BaseModel):
  model_config = ConfigDict(populate_by_name=True, extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  populate_by_name: This is to allow `meta` field to be populated by `_meta` or
  `meta`.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """

  inputs: Dict[str, Union[str, int, float, bool, APIWorkflowInConnection, dict]]
  class_type: str
  meta: Optional[APIWorkflowNodeMeta] = Field(None, alias='_meta')


class APIWorkflow(RootModel[Dict[APINodeID, APIWorkflowNodeInfo]]):
  """This is the API format, you get it from `Save (API Format)` in the UI.


  See test_data/sdxlturbo_example_api.json for an example of this format in json.
  """
  root: Dict[APINodeID, APIWorkflowNodeInfo]


################################################################################
class APISystemStatsSystem(BaseModel):
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """

  os: Optional[str] = None
  python_version: Optional[str] = None
  embedded_python: Optional[bool] = None


class APISystemStatsDevice(BaseModel):
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """
  name: Optional[str] = None
  type: Optional[str] = None
  index: Optional[int] = None
  vram_total: Optional[int] = None
  vram_free: Optional[int] = None
  torch_vram_total: Optional[int] = None
  torch_vram_free: Optional[int] = None


class APISystemStats(BaseModel):
  """Returned from /system_stats endpoint."""
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """
  system: Optional[APISystemStatsSystem] = None
  devices: Optional[List[APISystemStatsDevice]] = None


################################################################################
class APIPromptInfo(BaseModel):
  """Returned from /prompt endpoint."""

  class ExecInfo(BaseModel):
    queue_remaining: Optional[int]

  exec_info: Optional[ExecInfo]


################################################################################
class APIQueueInfoEntry(NamedTuple):
  number: int
  prompt_id: PromptID
  prompt: APIWorkflow
  extra_data: dict
  outputs_to_execute: List[APINodeID]


class APIQueueInfo(BaseModel):
  """Returned from /queue endpoint."""
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """

  queue_pending: List[APIQueueInfoEntry]
  queue_running: List[APIQueueInfoEntry]


################################################################################
class NodeErrorInfo(BaseModel):
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """
  details: str
  extra_info: dict
  message: str
  type: str


class NodeErrors(BaseModel):
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """
  class_type: str
  dependent_outputs: List[APINodeID]
  errors: List[NodeErrorInfo]


class APIWorkflowTicket(BaseModel):
  """Return from post /prompt endpoint."""
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """
  node_errors: Optional[Dict[APINodeID, NodeErrors]] = None
  number: Optional[int] = None
  prompt_id: Optional[PromptID] = None
  error: Union[NodeErrorInfo, str, None] = None


################################################################################


class APIOutputUI(RootModel[Dict[OutputName, List[Any]]]):
  root: Dict[OutputName, List[Any]]


class APIHistoryEntryStatusNote(NamedTuple):
  name: str
  data: Any


class APIHistoryEntryStatus(BaseModel):
  """

  Example:
    "status": {
      "status_str": "success",
      "completed": true,
      "messages": [
        [
          "execution_start",
          { "prompt_id": "b1b64df6-9b2c-4a09-bd0e-b6c294702085" }
        ],
        [
          "execution_cached",
          { "nodes": [], "prompt_id": "b1b64df6-9b2c-4a09-bd0e-b6c294702085" }
        ]
      ]
    }
  """
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """
  status_str: Optional[str] = None
  completed: Optional[bool] = None
  messages: Optional[List[APIHistoryEntryStatusNote]] = None


class APIHistoryEntry(BaseModel):
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """
  outputs: Optional[Dict[APINodeID, APIOutputUI]] = None
  prompt: Optional[APIQueueInfoEntry] = None
  status: Optional[APIHistoryEntryStatus] = None


class APIHistory(RootModel[Dict[PromptID, APIHistoryEntry]]):
  """Returned if you call /history and /history/{prompt_id} endpoints.

  TODO: Show an example.
  """
  root: Dict[PromptID, APIHistoryEntry]


################################################################################

APIObjectKey = Annotated[str, Field(alias='object_key')]
"""
  Example:

  KSampler:
    input:
      required:
        model:
        - MODEL
    ...

  Here 'KSampler' is the APIObjectKey.

"""


class APIObjectInputInfo(BaseModel):
  """
  Example:

  seed:
  - INT
  - default: 0
    min: 0
    max: 18446744073709551615
  """
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.

  I allow extra here because I don't know what the keys are, and they seem to
  vary quite a bit.
  """
  default: Optional[Any] = None
  min: Optional[Any] = None
  max: Optional[Any] = None
  step: Optional[Any] = None
  round: Optional[Any] = None
  # Note: Everything else is going to be in the extra dict. Access it with the
  # `extra` attribute.


class APIObjectInputTuple(NamedTuple):
  """
  Example:

  seed:
  - INT
  - default: 0
    min: 0
    max: 18446744073709551615

    First item in the list/tuple is the type, second is the optional info.
  """
  type: Union[NamedInputType, ComboInputType]
  # For some reason, when type=='*', this is an empty string.
  info: Union[APIObjectInputInfo, str, None] = None


class APIObjectInput(BaseModel):
  """
  input:
      required:
        model:
        - MODEL
        seed:
        - INT
        - default: 0
          min: 0
          max: 18446744073709551615
        ...
  """
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """

  required: Optional[Dict[str, Union[APIObjectInputTuple,
                                     NamedInputType]]] = None
  """
  For some reason, when type=='*', it just shows the type without a
  `[type, {... limits}] tuple, so I allowed NamedInputType.
  """

  optional: Optional[Dict[str, Union[APIObjectInputTuple,
                                     NamedInputType]]] = None
  hidden: Optional[Dict[str, Union[APIObjectInputTuple, NamedInputType]]] = None


class APIObjectInfoEntry(BaseModel):
  """

  Example yaml version of this:

  input:
      required:
        model:
        - MODEL
        seed:
        - INT
        - default: 0
          min: 0
          max: 18446744073709551615
        ...
    output:
    - LATENT
    output_is_list:
    - false
    output_name:
    - LATENT
    name: KSampler
    display_name: KSampler
    description: ''
    category: sampling
    output_node: false
  """
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """
  input: APIObjectInput
  output: Union[OutputType, List[Union[OutputType, List[OutputType]]]]
  output_is_list: List[bool]
  output_name: Union[OutputName, List[OutputName]]
  name: str
  display_name: str
  description: str
  category: str
  output_node: bool


class APIObjectInfo(RootModel[Dict[APIObjectKey, APIObjectInfoEntry]]):
  """Returned from /object_info endpoint.

  See test_data/object_info.yml for an example of this format in yaml.

  Example APIObjectInfo key and value, from test_data/object_info.yml:
  KSampler:
    input:
      required:
        model:
        - MODEL
        seed:
        - INT
        - default: 0
          min: 0
          max: 18446744073709551615
        steps:
        - INT
        - default: 20
          min: 1
          max: 10000
        cfg:
        - FLOAT
        - default: 8.0
          min: 0.0
          max: 100.0
          step: 0.1
          round: 0.01
        sampler_name:
        - - euler
          - euler_ancestral
          - heun
          - heunpp2
          - dpm_2
          - dpm_2_ancestral
          - lms
          - dpm_fast
          - dpm_adaptive
          - dpmpp_2s_ancestral
          - dpmpp_sde
          - dpmpp_sde_gpu
          - dpmpp_2m
          - dpmpp_2m_sde
          - dpmpp_2m_sde_gpu
          - dpmpp_3m_sde
          - dpmpp_3m_sde_gpu
          - ddpm
          - lcm
          - ddim
          - uni_pc
          - uni_pc_bh2
        scheduler:
        - - normal
          - karras
          - exponential
          - sgm_uniform
          - simple
          - ddim_uniform
        positive:
        - CONDITIONING
        negative:
        - CONDITIONING
        latent_image:
        - LATENT
        denoise:
        - FLOAT
        - default: 1.0
          min: 0.0
          max: 1.0
          step: 0.01
    output:
    - LATENT
    output_is_list:
    - false
    output_name:
    - LATENT
    name: KSampler
    display_name: KSampler
    description: ''
    category: sampling
    output_node: false
"""
  # model_config = ConfigDict(extra=EXTRA)

  root: Dict[APIObjectKey, APIObjectInfoEntry]


################################################################################
class APIUploadImageResp(BaseModel):
  name: str
  subfolder: str
  type: ComfyFolderType


################################################################################
class WSExecutingData(BaseModel):
  """Websocket "executing" message.
  See:

  * https://github.com/comfyanonymous/ComfyUI/blob/61b3f15f8f2bc0822cb98eac48742fb32f6af396/server.py#L115
  * https://github.com/comfyanonymous/ComfyUI/blob/c782144433e41c21ae2dfd75d0bc28255d2e966d/main.py#L113
  * https://github.com/comfyanonymous/ComfyUI/blob/c782144433e41c21ae2dfd75d0bc28255d2e966d/execution.py#L146
  * https://github.com/comfyanonymous/ComfyUI/blob/e478b1794e91977c50dc6eea6228ef1248044507/script_examples/websockets_api_example.py#L36


  """
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """
  node: Optional[str] = None
  prompt_id: Optional[PromptID] = None


class WSMessage(BaseModel):
  """Messages from the websocket, if it is non-binary."""
  model_config = ConfigDict(extra=EXTRA)
  """This is a pydantic thing, to configure the model, it is not an accessible field.

  extra: This is just to future proof the schema so it won't break if extra
  fields are added. They'll be stored dynamically.
  """
  type: str
  data: dict


################################################################################


class ComfyUIPathTriplet(BaseModel):
  """
  Represents a folder_type/subfolder/filename triplet, which ComfyUI API and
  some nodes use as file paths.
  """
  model_config = ConfigDict(frozen=True)

  type: ComfyFolderType
  subfolder: str
  filename: str

  @field_validator('type')
  @classmethod
  def validate_folder_type(cls, v: str):
    if v not in VALID_FOLDER_TYPES:
      raise ValueError(
          f'folder_type {json.dumps(v)} is not one of {VALID_FOLDER_TYPES}')
    return v

  @field_validator('subfolder')
  @classmethod
  def validate_subfolder(cls, v: str):
    if v.startswith('/'):
      raise ValueError(f'subfolder {json.dumps(v)} must not start with a slash')
    return v

  @field_validator('filename')
  @classmethod
  def validate_filename(cls, v: str):
    if '/' in v:
      raise ValueError(f'filename {json.dumps(v)} must not contain a slash')
    if v == '':
      raise ValueError(f'filename {json.dumps(v)} must not be empty')
    return v

  def ToLocalPathStr(self, *, include_folder_type: bool) -> str:
    """Converts this triplet to something like `input/subfolder/filename`.
    """
    subfolder = self.subfolder
    if subfolder == '':
      subfolder = '.'
    if not subfolder.endswith('/'):
      subfolder += '/'

    local_path = urljoin(subfolder, self.filename)
    if include_folder_type:
      local_path = urljoin(f'{self.type}/', local_path)
    return local_path



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import asyncio
import datetime
import enum
import json
import logging
import textwrap
import threading
import traceback as tb
import uuid
from copy import deepcopy
from dataclasses import dataclass
from pprint import pformat
from typing import (Any, Dict, Generic, List, Optional, Sequence, Tuple,
                    TypeVar, Union, overload)
from urllib.parse import urlparse

import aiofiles
from anyio import Path
from slugify import slugify
from typing_extensions import Literal
from websockets import WebSocketClientProtocol, connect

from ._internal.utilities import (BasicAuthToHeaders, DumpModelToDict,
                                  DumpModelToYAML, DumpYaml, GetWebSocketURL,
                                  TryParseAsModel)
from .api_client import ComfyAPIClientBase
from .catapult_base import (ComfyCatapultBase, ExceptionInfo, JobID, JobStatus,
                            Progress)
from .comfy_schema import (APIHistory, APIHistoryEntry,
                           APIHistoryEntryStatusNote, APINodeID, APIPromptInfo,
                           APIQueueInfo, APISystemStats, APIWorkflowTicket,
                           PromptID, WSMessage)
from .errors import (JobFailed, JobNotFound, NodesNotExecuted,
                     WorkflowSubmissionError)

logger = logging.getLogger(__name__)

_JobQueueStatus = Literal['pending', 'running', 'not_in_queue']


@dataclass
class _Job:

  class RemoteStatus(enum.Enum):
    NONE = enum.auto()
    PENDING_OR_RUNNING = enum.auto()

  job_id: JobID
  prepared_workflow: dict
  important: Tuple[APINodeID, ...]
  status: JobStatus
  # Exceptions that should be in status but aren't because they're not pickable
  # or deepcopyable.
  errors: List[Exception]
  future: 'asyncio.Future[dict]'
  remote_job_status: RemoteStatus
  job_debug_path: Optional[Path]


T = TypeVar('T')


@dataclass
class _Guess(Generic[T]):
  value: Optional[T]
  updated: datetime.datetime


class ComfyCatapult(ComfyCatapultBase):

  def __init__(self,
               comfy_client: ComfyAPIClientBase,
               *,
               debug_path: Optional[Path],
               debug_save_all: bool = False):
    """_summary_

    Args:
        comfy_client (ComfyAPIClientBase): _description_
        debug_path (Path | None): For logging of certain debug information, in
          case of errors.
        debug_save_all (bool, optional): If set, as much information as possible
          will be saved to the debug_path. Defaults to False.
    """
    self._comfy_client = comfy_client
    ############################################################################
    self._client_id = str(uuid.uuid4())

    # self._lock = _DummyLock()
    self._lock = asyncio.Lock()
    self._stop_event = threading.Event()

    self._jobs: Dict[JobID, _Job] = {}
    # {ticket.prompt_id -> job_id}
    self._prompt_id_index: Dict[PromptID, JobID] = {}

    self._monitoring_task: asyncio.Task[None] = asyncio.create_task(
        self._MonitoringThread())
    self._monitoring_task.get_loop().set_debug(True)
    self._poll_task: asyncio.Task[None] = asyncio.create_task(self._PollLoop())
    self._guess_currently_running_job_id: _Guess[str] = _Guess(
        value=None, updated=self._Now())
    self._guess_currently_running_node_id: _Guess[str] = _Guess(
        value=None, updated=self._Now())
    self._guess_currently_running_node_progress: _Guess[Progress] = _Guess(
        value=None, updated=self._Now())

    # Reconnect to the webscoket every 20 seconds, because the currently running
    # node is sent upon reconnect.
    self._ws_connect_interval = 20.
    self._loop_delay: float = 2.
    self._debug_path = debug_path
    self._debug_save_all = debug_save_all

  async def __aenter__(self):
    await self._comfy_client.__aenter__()
    return self

  async def __aexit__(self, exc_type, exc_value, traceback):
    await self.Close()

  async def Close(self):

    self._stop_event.set()

    async with self._lock:
      for job in self._jobs.values():
        job.future.cancel()
    await self._CheckError()
    async with self._lock:
      monitoring_task = self._monitoring_task
      poll_task = self._poll_task
    monitoring_task.cancel()
    poll_task.cancel()
    try:
      await monitoring_task
    except asyncio.CancelledError:
      pass
    try:
      await poll_task
    except asyncio.CancelledError:
      pass

  async def Resume(self,
                   *,
                   job_id: JobID,
                   prepared_workflow: dict,
                   important: Sequence[APINodeID],
                   status: JobStatus,
                   poll: bool,
                   errors: List[Exception] = [],
                   job_debug_path: Optional[Path] = None):
    if status.ticket is None:
      raise ValueError('status.ticket is None')
    if status.ticket.prompt_id is None:
      raise ValueError('status.ticket.prompt_id is None')
    job = _Job(job_id=job_id,
               prepared_workflow=prepared_workflow,
               important=tuple(important),
               status=status,
               errors=errors,
               future=asyncio.Future(),
               remote_job_status=_Job.RemoteStatus.PENDING_OR_RUNNING,
               job_debug_path=job_debug_path)

    prompt_id = status.ticket.prompt_id
    async with self._lock:
      if job_id in self._jobs:
        raise KeyError(f'User job id {json.dumps(job_id)} already exists')
      if prompt_id in self._prompt_id_index:
        raise KeyError(f'Prompt id {json.dumps(prompt_id)} already exists')

      self._jobs[job_id] = job
      self._prompt_id_index[prompt_id] = job_id
    if poll:
      await self._PollJobs(job_ids=[job_id])

  @overload
  async def Catapult(
      self,
      *,
      job_id: JobID,
      prepared_workflow: dict,
      important: Sequence[APINodeID],
      use_future_api: Literal[True],
      job_debug_path: Optional[Path] = None
  ) -> Tuple[JobStatus, 'asyncio.Future[dict]']:
    ...

  @overload
  async def Catapult(self,
                     *,
                     job_id: JobID,
                     prepared_workflow: dict,
                     important: Sequence[APINodeID],
                     use_future_api: Literal[False] = False,
                     job_debug_path: Optional[Path] = None) -> dict:
    ...

  async def Catapult(
      self,
      *,
      job_id: JobID,
      prepared_workflow: dict,
      important: Sequence[APINodeID],
      use_future_api: bool = False,
      job_debug_path: Optional[Path] = None,
  ) -> Union[dict, Tuple[JobStatus, 'asyncio.Future[dict]']]:
    # TODO: Deprecate use_future_api==False at version>=3.0.
    status, future = await self._CatapultInternal(
        job_id=job_id,
        prepared_workflow=prepared_workflow,
        important=important,
        job_debug_path=job_debug_path)
    if use_future_api:
      return status, future
    return await future

  async def _CatapultInternal(
      self,
      *,
      job_id: JobID,
      prepared_workflow: dict,
      important: Sequence[APINodeID],
      job_debug_path: Optional[Path] = None,
  ) -> Tuple[JobStatus, 'asyncio.Future[dict]']:
    async with self._lock:
      if job_id in self._jobs:
        raise KeyError(f'User job id {json.dumps(job_id)} already exists')
      if not slugify(job_id) == job_id:
        raise ValueError(
            f'User job id {json.dumps(job_id)} is not slugified (e.g {json.dumps(slugify(job_id))})'
        )
      if job_debug_path is None and self._debug_path is not None:
        dt = datetime.datetime.now(datetime.timezone.utc).isoformat()
        job_debug_path = self._debug_path / f'{slugify(dt)}-{slugify(job_id)}'

    if job_debug_path is not None:
      await job_debug_path.mkdir(parents=True, exist_ok=True)

    future: asyncio.Future[dict] = asyncio.Future()
    job = _Job(job_id=job_id,
               prepared_workflow=prepared_workflow,
               important=tuple(important),
               future=future,
               status=JobStatus(scheduled=self._Now(),
                                comfy_scheduled=None,
                                running=None,
                                pending=None,
                                success=None,
                                errored=None,
                                cancelled=None,
                                system_stats_check=None,
                                queue_check=None,
                                ticket=None,
                                job_history=None,
                                errors=[]),
               errors=[],
               remote_job_status=_Job.RemoteStatus.PENDING_OR_RUNNING,
               job_debug_path=job_debug_path)

    async with self._lock:
      self._jobs[job_id] = job

    async with _JobContext(job_id=job_id, catapult=self):
      ticket: APIWorkflowTicket = await self._comfy_client.PostPrompt(
          prompt_workflow=prepared_workflow)

      async with self._lock:
        job.status = job.status._replace(ticket=ticket)
        if ticket.prompt_id is not None:
          self._prompt_id_index[ticket.prompt_id] = job_id

      has_errors = False
      if ticket.node_errors is not None and len(ticket.node_errors) > 0:
        has_errors = True
      if ticket.error is not None:
        has_errors = True
      if ticket.prompt_id is None:
        has_errors = True

      if has_errors:
        prepared_workflow_yaml_str = await DumpYaml(prepared_workflow)
        ticket_yaml_str = await DumpModelToYAML(ticket)

        raise WorkflowSubmissionError(
            f'Errors in workflow'
            f'\nprepared_workflow:\n{textwrap.indent(prepared_workflow_yaml_str, prefix="  ")}'
            f'\nticket:\n{textwrap.indent(ticket_yaml_str, prefix="  ")}',
            prepared_workflow=deepcopy(prepared_workflow),
            ticket=ticket)

      if ticket.prompt_id is None:
        raise AssertionError(
            'ticket.prompt_id is None, should be impossible here')
      async with self._lock:
        job.status = job.status._replace(comfy_scheduled=self._Now())

      async with self._lock:
        return deepcopy(job.status), job.future

  async def GetStatus(
      self,
      *,
      job_id: JobID,
      poll: bool = False) -> 'Tuple[JobStatus, asyncio.Future[dict]]':
    if poll:
      await self._PollJobs(job_ids=[job_id])
    async with self._lock:
      job = self._GetJob(job_id=job_id)
      return deepcopy(job.status), job.future

  async def GetExceptions(self, *, job_id: JobID) -> List[Exception]:
    async with self._lock:
      job = self._GetJob(job_id=job_id)
      return list(job.errors)

  def _GetJob(self, *, job_id: JobID) -> _Job:
    """Get a job by job_id, raising JobNotFound if it doesn't exist.

    Must be called within a lock.
    """
    if self._lock.locked() is False:
      raise AssertionError('Must be called within a lock.')
    if job_id not in self._jobs:
      raise JobNotFound(f'Job id {json.dumps(job_id)} not found')
    return self._jobs[job_id]

  async def CancelJob(self, *, job_id: JobID):
    async with _JobContext(job_id=job_id, catapult=self):
      # Update job, for prompt id etc.
      await self._PollJobs(job_ids=[job_id])
      job_id_2_status: Dict[JobID, _JobQueueStatus]
      # Get the queue status of the job.
      job_id_2_status = await self._PollQueue(job_ids=[job_id])
      queue_status = job_id_2_status.get(job_id, 'not_in_queue')
      async with self._lock:
        job: _Job = self._GetJob(job_id=job_id)
        ticket: Optional[APIWorkflowTicket] = job.status.ticket
        prompt_id: Optional[PromptID] = None
        if ticket is not None:
          prompt_id = ticket.prompt_id

      if queue_status == 'running' \
          and prompt_id is not None:
        # TODO(realazthat/comfy-catapult#5): We don't know for sure that the
        # current job is the one we're cancelling, as there could be a race
        # condition here.
        await self._comfy_client.PostInterrupt()
        async with self._lock:
          job.remote_job_status = _Job.RemoteStatus.NONE
      elif queue_status == 'pending' and prompt_id is not None:
        await self._comfy_client.PostQueue(delete=[prompt_id], clear=False)

        async with self._lock:
          job.remote_job_status = _Job.RemoteStatus.NONE

      now = self._Now()
      async with self._lock:
        if not job.status.IsDone():
          job.status = job.status._replace(cancelled=now)
          job.future.cancel()

  async def _ReceivedJobHistory(self, *, job_id: JobID, history: APIHistory,
                                queue_status: _JobQueueStatus,
                                job_context: '_JobContext'):
    if not isinstance(history, APIHistory):
      raise AssertionError(f'history must be APIHistory, not {type(history)}')

    async with self._lock:
      job: _Job = self._GetJob(job_id=job_id)

    async with self._lock:
      prepared_workflow: dict = deepcopy(job.prepared_workflow)
      ticket: Optional[APIWorkflowTicket] = deepcopy(job.status.ticket)
      prompt_id: Optional[PromptID] = None
      if ticket is not None:
        prompt_id = ticket.prompt_id
      important = deepcopy(job.important)

    if len(history.root) == 0:
      if queue_status == 'not_in_queue':
        raise Exception(
            'Job disappeared from the ComfyUI. Perhaps ComfyUI was'
            ' restarted, due to a very bad crash, e.g a GPU crash or'
            ' a segfault.'
            f'\n  job_id: {json.dumps(job_id)}'
            f'\n  prompt_id: {json.dumps(prompt_id)}')
      return
    if prompt_id not in history.root:
      raise AssertionError(
          f'prompt_id {json.dumps(prompt_id)} not in history.root')
    if not isinstance(prompt_id, str):
      raise AssertionError(f'prompt_id must be str, not {type(prompt_id)}')
    job_history: APIHistoryEntry = history.root[prompt_id]
    async with self._lock:
      job.status = job.status._replace(
          job_history=await DumpModelToDict(job_history))
    ##########################################################################
    # Get outputs_to_execute, outputs_with_data, extra_data
    extra_data: Optional[dict] = None
    outputs_to_execute: List[APINodeID] = []
    outputs_with_data: List[APINodeID] = []
    if job_history.outputs is not None:
      outputs_with_data = list(job_history.outputs.keys())
    if job_history.prompt is not None:
      if job_history.prompt.extra_data is not None:
        extra_data = job_history.prompt.extra_data
      if job_history.prompt.outputs_to_execute is not None:
        outputs_to_execute = job_history.prompt.outputs_to_execute
    logger.debug('extra_data: %s', await DumpYaml(extra_data))

    if job_history.status is not None:
      if job_history.status.completed is False:
        notes: List[str] = []
        if job_history.status.messages is not None:
          note: APIHistoryEntryStatusNote
          for note in job_history.status.messages:
            notes.append(textwrap.indent(pformat(note._asdict()), prefix='  '))
        # TODO: Turn all Exception into a subclass of Exception.
        # TODO: Make all exceptions going forward contain the metadata that
        # NodesNotExecuted does.
        raise JobFailed('Job has failed'
                        f'\n  status: {repr(job_history.status)}'
                        f'\n  notes:' + '\n'.join(notes))

    ##########################################################################
    logger.debug('job_history.prompt.extra_data: %s', await
                 DumpYaml(extra_data))
    logger.debug('job_history.prompt.outputs_to_execute: %s', await
                 DumpYaml(outputs_to_execute))

    logger.debug('outputs_to_execute: %s', await DumpYaml(outputs_to_execute))
    logger.debug('outputs_that_executed: %s', await DumpYaml(outputs_with_data))

    bad_dataless_outputs = [
        node_id for node_id in outputs_to_execute
        if node_id not in outputs_with_data
    ]

    def _GetTitles(node_ids: List[APINodeID]) -> List[Optional[str]]:
      titles = []
      for node_id in node_ids:
        node_info: dict = prepared_workflow.get(node_id, {})
        meta: dict = node_info.get('_meta', {})
        title: Optional[str] = meta.get('title', None)
        titles.append(title)
      return titles

    if len(bad_dataless_outputs) > 0:
      logger.error('bad_dataless_outputs: %s', await
                   DumpYaml(bad_dataless_outputs))
      logger.error('_GetTitles(bad_dataless_outputs): %s', await
                   DumpYaml(_GetTitles(bad_dataless_outputs)))

    dataless_important_outputs = [
        node_id for node_id in bad_dataless_outputs if node_id in important
    ]
    if len(dataless_important_outputs) > 0:
      logger.error('dataless_important_outputs: %s', await
                   DumpYaml(dataless_important_outputs))
      logger.error('_GetTitles(dataless_important_outputs): %s', await
                   DumpYaml(_GetTitles(dataless_important_outputs)))
      raise NodesNotExecuted(nodes=list(dataless_important_outputs),
                             titles=_GetTitles(dataless_important_outputs))

    # await self._eta_estimator.RecordFinished(job_id=job_id)

    now = self._Now()
    async with self._lock:
      if job.status.success is None:
        job.status = job.status._replace(success=now)
      if not job.future.done():
        job.future.set_result(await DumpModelToDict(job_history))

  async def _GetJobIDs(self) -> List[JobID]:
    async with self._lock:
      return list(self._jobs.keys())

  async def _PollJobs(self, *, job_ids: List[JobID]):
    await self._PollSystemStats(job_ids=job_ids)
    job_id_2_status = await self._PollQueue(job_ids=job_ids)
    # Update remote status based on /queue endpoint.
    await self._PollUpdateRemoteJobStatus(job_ids=job_ids,
                                          job_id_2_status=job_id_2_status)
    # Check the /history endpoint to see if there are any updates on our jobs.
    await self._PollHistory(job_id_2_status=job_id_2_status)
    await self._PollFutures(job_ids=job_ids)

  async def _PollOnce(self):
    ############################################################################
    job_ids = await self._GetJobIDs()
    ############################################################################
    prompt_info: APIPromptInfo = await self._comfy_client.GetPrompt()

    exec_info: Optional[APIPromptInfo.ExecInfo] = prompt_info.exec_info
    queue_remaining: Optional[int] = None
    if exec_info is None:
      logger.warning('/prompt endpoint has no exec_info')
    else:
      queue_remaining = exec_info.queue_remaining
    if queue_remaining is not None:
      logger.info('queue_remaining: %s', queue_remaining)
    else:
      logger.warning('queue_remaining: Unavailable, ComfyUI protocol changed.')
    ############################################################################
    await self._PollJobs(job_ids=job_ids)

  async def _PollFutures(self, *, job_ids: List[JobID]):
    ############################################################################
    # Make sure any hanging future that is done matches the job status.
    done_futures_jobs: List[JobID] = []
    async with self._lock:
      job_id: JobID
      job: _Job
      for job_id in job_ids:
        try:
          async with _JobContext(job_id=job_id, catapult=self):
            job = self._GetJob(job_id=job_id)
            if job.future.done() and not job.status.IsDone():
              done_futures_jobs.append(job_id)
        except KeyError:
          logger.exception(
              f'Error in _PollFutures for job_id {json.dumps(job_id)}. Continuing.'
          )
    for job_id in done_futures_jobs:
      try:
        async with _JobContext(job_id=job_id, catapult=self):
          async with self._lock:
            job = self._GetJob(job_id=job_id)
            # See if an error occurred.
            job.future.result()
      except Exception:
        logger.exception(
            f'Error in _PollFutures for job_id {json.dumps(job_id)}. Continuing.'
        )

  async def _PollSystemStats(self, *, job_ids: List[JobID]):
    system_stats: APISystemStats = await self._comfy_client.GetSystemStats()
    logger.info('system_stats: %s', await DumpModelToYAML(system_stats))
    for job_id in job_ids:
      try:
        async with _JobContext(job_id=job_id, catapult=self):
          async with self._lock:
            job = self._GetJob(job_id=job_id)
            if job.status.IsDone():
              continue
            if job.status.system_stats_check is None:
              job.status = job.status._replace(system_stats_check=self._Now())
      except Exception:
        logger.exception(
            f'Error in _PollFutures for job_id {json.dumps(job_id)}. Continuing.'
        )

  async def _PollQueue(self, *,
                       job_ids: List[JobID]) -> Dict[JobID, _JobQueueStatus]:
    """Check /queue endpoint and update job status.
    """

    queue_info: APIQueueInfo = await self._comfy_client.GetQueue()
    prompt_id_2_status: Dict[PromptID, _JobQueueStatus] = {}
    for pending in queue_info.queue_running:
      prompt_id_2_status[pending.prompt_id] = 'running'
    for running in queue_info.queue_pending:
      prompt_id_2_status[running.prompt_id] = 'pending'
    pending_count = sum(
        [1 for status in prompt_id_2_status.values() if status == 'pending'],
        start=0)
    running_count = sum(
        [1 for status in prompt_id_2_status.values() if status == 'running'],
        start=0)

    logger.info('pending_count: %s', pending_count)
    logger.info('running_count: %s', running_count)

    job_id_2_status: Dict[JobID, _JobQueueStatus] = {}
    for prompt_id, status in prompt_id_2_status.items():
      async with self._lock:
        if prompt_id not in self._prompt_id_index:
          continue
        job_id = self._prompt_id_index[prompt_id]
        if job_id not in job_ids:
          continue
        job_id_2_status[job_id] = status
    for job_id in job_ids:
      if job_id not in job_id_2_status:
        job_id_2_status[job_id] = 'not_in_queue'

    for job_id, status in job_id_2_status.items():
      try:
        async with _JobContext(job_id=job_id, catapult=self):
          async with self._lock:
            job: _Job = self._GetJob(job_id=job_id)

            if status == 'pending' and job.status.pending is None:
              job.status = job.status._replace(pending=self._Now())

            if status == 'running' and job.status.running is None:
              job.status = job.status._replace(running=self._Now())
              self._guess_currently_running_job_id = _Guess(value=job_id,
                                                            updated=self._Now())
              self._guess_currently_running_node_id = _Guess(
                  value=None, updated=self._Now())
              self._guess_currently_running_node_progress = _Guess(
                  value=None, updated=self._Now())
            job.status = job.status._replace(queue_check=self._Now())
      except Exception:
        logger.exception(
            f'Error in _PollQueue for job_id {json.dumps(job_id)}. Continuing.')

    # TODO: Reenable this or remove it.
    # print('self._jobs:', file=sys.stderr)
    # pprint(self._jobs, indent=2, stream=sys.stderr)
    return job_id_2_status

  async def _PollUpdateRemoteJobStatus(self, *, job_ids: List[JobID],
                                       job_id_2_status: Dict[JobID,
                                                             _JobQueueStatus]):
    for job_id in job_ids:
      try:
        async with _JobContext(job_id=job_id, catapult=self):
          queue_status = job_id_2_status.get(job_id, 'not_in_queue')
          async with self._lock:
            job = self._GetJob(job_id=job_id)
            new_remote_job_status = (_Job.RemoteStatus.NONE
                                     if queue_status == 'not_in_queue' else
                                     _Job.RemoteStatus.PENDING_OR_RUNNING)
            job.remote_job_status = new_remote_job_status
      except Exception:
        logger.exception(
            f'Error in _PollHistoryUpdateRemoteJobStatus for job_id {json.dumps(job_id)}. Continuing.'
        )

  async def _PollHistoryUpdateJobHistory(self, *, job_id: JobID,
                                         queue_status: _JobQueueStatus,
                                         job_context: '_JobContext'):
    async with self._lock:
      job: _Job = self._GetJob(job_id=job_id)

      if job.status.job_history is not None:
        # We already have the job history.
        return

      ticket: Optional[APIWorkflowTicket] = job.status.ticket
      prompt_id: Optional[PromptID] = None
      if ticket is not None:
        prompt_id = ticket.prompt_id

      if prompt_id is None:
        # This should never happen, but :shrug:.
        raise Exception('prompt_id is None; this should never happen, because'
                        ' we should have a ticket as soon as we submit a job.')
      job_debug_path = job.job_debug_path
      errors_dump_directory: Optional[Path] = None
      if job_debug_path is not None:
        errors_dump_directory = job_debug_path / 'errors'

    ########################################################################
    history_raw: dict = await self._comfy_client.GetHistoryRaw(
        prompt_id=prompt_id)
    await job_context.WatchVar(history_raw=history_raw)
    history: APIHistory = await TryParseAsModel(
        content=history_raw,
        model_type=APIHistory,
        errors_dump_directory=errors_dump_directory)

    await self._ReceivedJobHistory(job_id=job_id,
                                   history=history,
                                   queue_status=queue_status,
                                   job_context=job_context)
    ########################################################################

  async def _PollHistory(self, job_id_2_status: Dict[JobID, _JobQueueStatus]):
    # Update job.remote_job_status.
    for job_id in job_id_2_status.keys():
      try:
        async with _JobContext(job_id=job_id, catapult=self) as job_context:
          queue_status = job_id_2_status.get(job_id, 'not_in_queue')

          if queue_status in ['pending', 'running']:
            # In this case, it's pending or running, not going to be in the
            # `/history` endpoint.
            return
          await self._PollHistoryUpdateJobHistory(job_id=job_id,
                                                  queue_status=queue_status,
                                                  job_context=job_context)
      except Exception:
        logger.exception(
            f'Error in _PollHistory for job_id {json.dumps(job_id)}. Continuing.'
        )

  def _Now(self) -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)

  async def _Record(self):
    async with self._lock:
      guess_currently_running_job_id = self._guess_currently_running_job_id
      # guess_currently_running_node_id = self._guess_currently_running_node_id
      # guess_currently_running_node_progress = self._guess_currently_running_node_progress

    if guess_currently_running_job_id.value is not None:
      pass

  async def _LoopWS(self, *, ws: WebSocketClientProtocol):
    prompt_id: Optional[PromptID]
    node_id: Optional[str]
    node_type: Optional[str]

    # Only sleep if there is an error to prevent overflow to the logs; otherwise
    # poll ws.recv() as fast as possible.
    while True:
      try:
        logger.debug('websocket recv')
        out = await ws.recv()
        if not isinstance(out, str):
          logger.debug('websocket type(out): %s', type(out))
          continue
        logger.debug('websocket raw: %s', await DumpYaml(out))
        errors_dump_directory: Optional[Path] = None
        async with self._lock:
          if self._debug_path is not None:
            errors_dump_directory = self._debug_path / 'errors'

        message = await TryParseAsModel(
            content=json.loads(out),
            model_type=WSMessage,
            errors_dump_directory=errors_dump_directory)
        logger.debug('websocket message: %s', await DumpYaml(message.__dict__))

        if message.type == 'executing' and 'last_node_id' in message.data:
          last_node_id = message.data['last_node_id']

          async with self._lock:
            self._guess_currently_running_node_id = _Guess(value=last_node_id,
                                                           updated=self._Now())
          await self._Record()
        elif message.type == 'executing' and 'node' in message.data and 'prompt_id' in message.data:
          prompt_id = message.data.get('prompt_id', None)
          node_id = message.data.get('node', None)
          async with self._lock:
            job_id = self._prompt_id_index.get(
                prompt_id, None) if prompt_id is not None else None
            if job_id is not None:
              job = self._GetJob(job_id=job_id)
              if job.status.running is None:
                job.status = job.status._replace(running=self._Now())
            self._guess_currently_running_job_id = _Guess(value=job_id,
                                                          updated=self._Now())
            self._guess_currently_running_node_id = _Guess(value=node_id,
                                                           updated=self._Now())
            self._guess_currently_running_node_progress = _Guess(
                value=None, updated=self._Now())
          await self._Record()
        elif message.type == 'execution_start' and 'prompt_id' in message.data:
          prompt_id = message.data['prompt_id']
          async with self._lock:
            job_id = self._prompt_id_index.get(
                prompt_id, None) if prompt_id is not None else None
            if job_id is not None:
              job = self._GetJob(job_id=job_id)
              if job.status.running is None:
                job.status = job.status._replace(running=self._Now())
            self._guess_currently_running_job_id = _Guess(value=job_id,
                                                          updated=self._Now())
            self._guess_currently_running_node_id = _Guess(value=None,
                                                           updated=self._Now())
            self._guess_currently_running_node_progress = _Guess(
                value=None, updated=self._Now())
          await self._Record()

        elif message.type == 'executed' and 'node' in message.data and 'output_ui' in message.data and 'prompt_id' in message.data:
          # Finished a single node?
          prompt_id = message.data.get('prompt_id', None)
          node_id = message.data.get('node', None)
          # output_ui = message.data['output_ui']
          async with self._lock:
            job_id = self._prompt_id_index.get(
                prompt_id, None) if prompt_id is not None else None
            if job_id is not None:
              job = self._GetJob(job_id=job_id)
              if job.status.running is None:
                job.status = job.status._replace(running=self._Now())
            else:
              node_id = None

            self._guess_currently_running_job_id = _Guess(value=job_id,
                                                          updated=self._Now())
            self._guess_currently_running_node_id = _Guess(value=node_id,
                                                           updated=self._Now())
            self._guess_currently_running_node_progress = _Guess(
                value=None, updated=self._Now())
          await self._Record()
        elif message.type in ['execution_interrupted', 'execution_error']:
          prompt_id = message.data.get('prompt_id', None)
          node_id = message.data.get('node_id', None)
          node_type = message.data.get('node_type', None)
          # executed: Optional[list] = message.data.get('executed', None)
          async with self._lock:
            job_id = self._prompt_id_index.get(
                prompt_id, None) if prompt_id is not None else None
            if job_id is not None:
              job = self._GetJob(job_id=job_id)
              if not job.status.IsDone():
                now = self._Now()

                message_data_yaml_str = await DumpYaml(message.data)
                job.status = job.status._replace(
                    errored=now,
                    errors=job.status.errors + [
                        ExceptionInfo(
                            type=str(node_type),
                            message=
                            f'Node {json.dumps(node_id)} of type {json.dumps(str(node_type))} errored:\n{textwrap.indent(message_data_yaml_str, "  ")}',
                            traceback='',
                            attributes={})
                    ])
            self._guess_currently_running_job_id = _Guess(value=None,
                                                          updated=self._Now())
            self._guess_currently_running_node_id = _Guess(value=None,
                                                           updated=self._Now())
            self._guess_currently_running_node_progress = _Guess(
                value=None, updated=self._Now())
          await self._Record()
        elif message.type == 'progress':
          value = message.data.get('value', None)
          max_value = message.data.get('max_value', None)
          if isinstance(value, int) and isinstance(max_value, int):
            async with self._lock:
              self._guess_currently_running_node_progress = _Guess(
                  value=Progress(value=value, max_value=max_value),
                  updated=self._Now())
            await self._Record()
      except asyncio.CancelledError:
        raise
      except Exception:
        logger.exception('Error in _LoopWS')
        await asyncio.sleep(self._loop_delay)

  async def _PollLoop(self):
    while not self._stop_event.is_set():
      try:
        await self._PollOnce()
      except asyncio.CancelledError:
        raise
      except Exception:
        logger.exception('Error in _PollLoop')
      await asyncio.sleep(self._loop_delay)

  async def _MonitoringThread(self):

    async with self._lock:
      client_id = self._client_id
      comfy_api_url: str = self._comfy_client.GetURL()
      ws_connect_interval: float = self._ws_connect_interval

    ws_url = urlparse(
        GetWebSocketURL(comfy_api_url=comfy_api_url, client_id=client_id))

    ws_headers: Dict[str, str] = {}
    ws_url = urlparse(
        BasicAuthToHeaders(url=ws_url.geturl(), headers=ws_headers))

    async def _MonitoringThreadLoopOnce():
      nonlocal ws_url
      try:
        # websockets lib doessn't support basic auth in the url, so we have to
        # move it to the headers.

        async with connect(ws_url.geturl(), extra_headers=ws_headers) as ws:
          await asyncio.wait_for(self._LoopWS(ws=ws),
                                 timeout=ws_connect_interval)
      except asyncio.TimeoutError:
        pass
      except asyncio.CancelledError:
        raise
      except Exception:
        logger.exception('Error in _MonitoringThreadLoopOnce')

    while not self._stop_event.is_set():
      try:
        await _MonitoringThreadLoopOnce()
      except asyncio.TimeoutError:
        pass
      except asyncio.CancelledError:
        raise
      except Exception:
        logger.exception('Error in _MonitoringThreadLoopOnce')
      await asyncio.sleep(self._loop_delay)

  async def _CheckError(self):
    async with self._lock:
      if self._monitoring_task.done():
        self._monitoring_task.result()
      if self._poll_task.done():
        self._poll_task.result()


class _DummyLock:

  def acquire(self):
    pass

  def release(self):
    pass

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    pass


class _JobContext:
  """Useful for catching errors when doing internal operations on a job, and the setting the status appropriately.
  """

  def __init__(self, *, job_id: JobID, catapult: ComfyCatapult):
    self._job_id = job_id
    self._catapult = catapult
    self._watch_vars: Dict[str, Any] = {}

  async def WatchVar(self, **kwargs):
    for name, value in kwargs.items():
      self._watch_vars[name] = value
    async with self._catapult._lock:
      job = self._catapult._GetJob(job_id=self._job_id)
      job_debug_path: Optional[Path] = job.job_debug_path
    if job_debug_path is None:
      return
    if self._catapult._debug_save_all:
      dt = datetime.datetime.now(datetime.timezone.utc).isoformat()
      for name, value in kwargs.items():
        dump_path = job_debug_path / 'watch' / f'{slugify(dt)}-{slugify(name)}.yaml'
        await dump_path.parent.mkdir(parents=True, exist_ok=True)
        async with aiofiles.open(dump_path, 'w') as f:
          await f.write(await DumpYaml(value))

  async def __aenter__(self):
    return self

  async def __aexit__(self, exc_type, exc_value, traceback):
    if exc_type is not None:
      async with self._catapult._lock:
        if self._job_id not in self._catapult._jobs:
          return
        job = self._catapult._GetJob(job_id=self._job_id)
        if job.future is not None and not job.future.done():
          job.future.set_exception(exc_value)
        if not job.status.IsDone():
          job.errors += [exc_value]
          exc_info = ExceptionInfo(type=exc_type.__name__,
                                   message=str(exc_value),
                                   traceback=''.join(
                                       tb.format_exception(
                                           exc_type, exc_value, traceback)),
                                   attributes={
                                       k: str(v)
                                       for k, v in exc_value.__dict__.items()
                                   })
          job.status = job.status._replace(errored=self._catapult._Now(),
                                           errors=job.status.errors +
                                           [exc_info])
        job_debug_path: Optional[Path] = job.job_debug_path
        status = deepcopy(job.status)
        workflow = deepcopy(job.prepared_workflow)
        ticket = deepcopy(job.status.ticket)
        remote_job_status = deepcopy(job.remote_job_status)

      dt = datetime.datetime.now(datetime.timezone.utc).isoformat()
      if job_debug_path is not None:
        dump = {
            'job_id': self._job_id,
            'exc_type': exc_type,
            'exc_value': str(exc_value),
            'traceback': tb.format_exception(exc_type, exc_value, traceback),
            'job_status': status,
            'remote_job_status': remote_job_status,
            'ticket': ticket,
            'watch_vars': self._watch_vars,
            'workflow': workflow,
        }
        job_dump_path = job_debug_path / 'context-dumps' / f'{slugify(dt)}.dump.yaml'
        await job_dump_path.parent.mkdir(parents=True, exist_ok=True)
        async with aiofiles.open(job_dump_path, 'w') as f:
          await f.write(await DumpYaml(dump))
        logger.error(f'Wrote job status to {json.dumps(str(job_dump_path))}.')



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

from abc import ABC, abstractmethod
from typing import List, Optional

from .comfy_schema import (APIHistory, APIObjectInfo, APIPromptInfo,
                           APIQueueInfo, APISystemStats, APIUploadImageResp,
                           APIWorkflowTicket, ClientID, PromptID)


class ComfyAPIClientBase(ABC):

  @abstractmethod
  async def __aenter__(self):
    raise NotImplementedError()

  @abstractmethod
  async def __aexit__(self, exc_type, exc, tb):
    raise NotImplementedError()

  @abstractmethod
  def GetURL(self) -> str:
    raise NotImplementedError()

  @abstractmethod
  async def Close(self):
    raise NotImplementedError()

  @abstractmethod
  async def GetSystemStatsRaw(self) -> dict:
    raise NotImplementedError()

  @abstractmethod
  async def GetSystemStats(self) -> APISystemStats:
    raise NotImplementedError()

  @abstractmethod
  async def GetObjectInfoRaw(self) -> dict:
    """Returns all known nodes and their metadata.

    Gets the data from the `/object_info` endpoint.

    See test_data/object_info.yml for an example of the response.

    See APIObjectInfo documentation for more info.

    Returns:
        dict: The parsed json, as a python dict.
    """
    raise NotImplementedError()

  @abstractmethod
  async def GetObjectInfo(self) -> APIObjectInfo:
    """Returns all known nodes and their metadata.

    Gets the data from the `/object_info` endpoint.

    See APIObjectInfo documentation for more info.

    Returns:
        APIObjectInfo: The parsed response.
    """
    raise NotImplementedError()

  @abstractmethod
  async def GetPromptRaw(self) -> dict:
    raise NotImplementedError()

  @abstractmethod
  async def GetPrompt(self) -> APIPromptInfo:
    raise NotImplementedError()

  @abstractmethod
  async def PostPromptRaw(self,
                          *,
                          prompt_workflow: dict,
                          number: Optional[int] = None,
                          client_id: Optional[ClientID] = None,
                          prompt_id: Optional[PromptID] = None,
                          extra_data: Optional[dict] = None) -> dict:
    """See `ComfyUI/server.py` `@routes.post("/prompt")`.

    Args:
        prompt_workflow (dict): The API workflow, you can generate this with
          the ComfyUI web interface, as explained in `README.md`:
          1. Open settings (gear box in the corner).
          2. Enable the ability to export in the API format,
            `Enable Dev mode Options`.
          3. Click new menu item `Save (API format)`.
        number (int, optional): Where to insert into the queue. -1 means to
          insert to the front of the queue. Default is None, which uses the
          default on the server, which is also equivalent to -1.
        client_id (ClientID | None): If you want to set the client id. ComfyUI
          will use this to resume connections, and filter out websockets events
          and similar etc. If you leave this empty, the server will generate a
          client id for you.
        prompt_id (PromptID | None): If you want to set the prompt id. If you
          leave this empty, the server will generate a prompt id for you.
        extra_data (dict | None, optional): Extra data associated with the
          prompt/job. Defaults to None.

    Returns:
        dict: The json response.
    """
    raise NotImplementedError()

  @abstractmethod
  async def PostPrompt(self,
                       *,
                       prompt_workflow: dict,
                       number: Optional[int] = None,
                       client_id: Optional[ClientID] = None,
                       prompt_id: Optional[PromptID] = None,
                       extra_data: Optional[dict] = None) -> APIWorkflowTicket:
    """See `ComfyUI/server.py` `@routes.post("/prompt")`.

    Args:
        prompt_workflow (dict): The API workflow, you can generate this with
          the ComfyUI web interface, as explained in `README.md`:
          1. Open settings (gear box in the corner).
          2. Enable the ability to export in the API format,
            `Enable Dev mode Options`.
          3. Click new menu item `Save (API format)`.
        number (int, optional): Where to insert into the queue. -1 means to
          insert to the front of the queue. Default is None, which uses the
          default on the server, which is also equivalent to -1.
        client_id (ClientID | None): If you want to set the client id. ComfyUI
          will use this to resume connections, and filter out websockets events
          and similar etc. If you leave this empty, the server will generate a
          client id for you.
        prompt_id (PromptID | None): If you want to set the prompt id. If you
          leave this empty, the server will generate a prompt id for you.
        extra_data (dict | None, optional): Extra data associated with the
          prompt/job. Defaults to None.

    Returns:
        APIWorkflowTicket: The response parsed into a pydantic model.
    """
    raise NotImplementedError()

  @abstractmethod
  async def GetHistoryRaw(self,
                          *,
                          prompt_id: Optional[PromptID] = None,
                          max_items: Optional[int] = None) -> dict:
    raise NotImplementedError()

  @abstractmethod
  async def GetHistory(self,
                       *,
                       prompt_id: Optional[PromptID] = None,
                       max_items: Optional[int] = None) -> APIHistory:
    raise NotImplementedError()

  @abstractmethod
  async def GetQueueRaw(self) -> dict:
    raise NotImplementedError()

  @abstractmethod
  async def GetQueue(self) -> APIQueueInfo:
    raise NotImplementedError()

  @abstractmethod
  async def GetView(self, *, folder_type: str, subfolder: str,
                    filename: str) -> bytes:
    raise NotImplementedError()

  @abstractmethod
  async def PostUploadImageRaw(self, *, folder_type: str, subfolder: str,
                               filename: str, data: bytes,
                               overwrite: bool) -> dict:
    raise NotImplementedError()

  @abstractmethod
  async def PostUploadImage(self, *, folder_type: str, subfolder: str,
                            filename: str, data: bytes,
                            overwrite: bool) -> APIUploadImageResp:
    raise NotImplementedError()

  @abstractmethod
  async def PostFree(self, *, unload_models: bool, free_memory: bool):
    raise NotImplementedError()

  @abstractmethod
  async def PostInterrupt(self):
    """
    # TODO(realazthat/comfy-catapult#5): change the API to take a prompt_id.
    """
    raise NotImplementedError()

  @abstractmethod
  async def PostQueue(self, *, delete: List[PromptID], clear: bool):
    raise NotImplementedError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import asyncio
import base64
import dataclasses
import datetime
import functools
import json
import logging
import sys
import textwrap
import traceback
from dataclasses import is_dataclass
from typing import (Any, Callable, Dict, Generator, List, Literal, NamedTuple,
                    Optional, Type, TypeVar)
from urllib.parse import unquote as paramdecode
from urllib.parse import urlparse

import aiofiles
import pydantic_core
import yaml
from anyio import Path
from pydantic import BaseModel
from pydash import slugify

from .url_utils import JoinToBaseURL

MAX_DUMP_LINES: Optional[int] = 200
_BaseModelT = TypeVar('_BaseModelT', bound=BaseModel)
logger = logging.getLogger(__name__)

if sys.version_info >= (3, 9):
  from asyncio import to_thread
else:
  T = TypeVar('T')

  async def to_thread(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """Run a function in a separate thread."""
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None,
                                      functools.partial(func, *args, **kwargs))


class _CustomDumper(yaml.Dumper):

  def represent_tuple(self, data):
    return self.represent_list(data)


_CustomDumper.add_representer(tuple, _CustomDumper.represent_tuple)


async def DumpModelToDict(model: BaseModel, **kwargs) -> Dict[str, Any]:
  if 'by_alias' not in kwargs:
    kwargs['by_alias'] = True
  if 'round_trip' not in kwargs:
    kwargs['round_trip'] = True
  if 'mode' not in kwargs:
    kwargs['mode'] = 'json'
  return await to_thread(model.model_dump, **kwargs)


async def DumpYaml(data: Any) -> str:
  return await to_thread(yaml.dump,
                         data,
                         indent=2,
                         Dumper=_CustomDumper,
                         sort_keys=False)


async def DumpModelToYAML(model: BaseModel, **kwargs) -> str:
  model_dict = await DumpModelToDict(model)
  return await DumpYaml(model_dict)


async def DumpBigYaml(data: Any, *, max_lines: Optional[int],
                      path: Path) -> str:
  yaml_str = await DumpYaml(data)
  line_count = len(yaml_str.splitlines())
  if max_lines is None or line_count <= max_lines:
    return yaml_str

  await path.parent.mkdir(parents=True, exist_ok=True)
  async with aiofiles.open(path, 'w') as f:
    await f.write(yaml_str)
  return f'Too large, see {json.dumps(str(path))}'


async def DumpBigErrorStr(exception: Exception, max_lines: Optional[int],
                          path: Path) -> str:
  if max_lines is None:
    return str(exception)
  error_str = str(exception)
  error_lines = error_str.splitlines()
  error_line = error_lines[0]

  await path.parent.mkdir(parents=True, exist_ok=True)

  async with aiofiles.open(path, 'w') as f:
    await f.write(error_str)
  return error_line + f' (Wrote full error_str to {json.dumps(str(path))})'


def BasicAuthToHeaders(*, url: str, headers: Dict[str, str]) -> str:
  """
  websockets lib doessn't support basic auth in the url, so we have to move it
  to the headers.
  """
  url_pr = urlparse(url)
  if url_pr.username is None and url_pr.password is None:
    return url

  username = url_pr.username or ''
  password = url_pr.password or ''

  # urldecode the username and password
  username = paramdecode(username)
  password = paramdecode(password)
  auth = f'{username}:{password}'
  encoded_auth = base64.b64encode(auth.encode()).decode()

  headers['Authorization'] = f'Basic {encoded_auth}'
  new_netloc = f'{url_pr.hostname}:{url_pr.port}'
  return url_pr._replace(netloc=new_netloc).geturl()


def GetWebSocketURL(*, comfy_api_url: str, client_id: str) -> str:
  ws_url_str = JoinToBaseURL(comfy_api_url, 'ws')
  ws_url = urlparse(ws_url_str)
  if ws_url.scheme == 'https':
    ws_url = ws_url._replace(scheme='wss')
  else:
    ws_url = ws_url._replace(scheme='ws')
  ws_url = ws_url._replace(query=f'clientId={client_id}')
  return ws_url.geturl()


def _IsDataclassInstance(instance):
  return is_dataclass(instance) and not isinstance(instance, type)


class _ExtraFieldWarning(NamedTuple):
  path: List[Any]
  thing: Any
  message: str


def _WarnModelExtras(*, path: List[Any],
                     thing: Any) -> Generator[_ExtraFieldWarning, None, None]:
  if _IsDataclassInstance(thing):
    for field in dataclasses.fields(thing):
      item = getattr(thing, field.name)
      yield from _WarnModelExtras(path=path + [field.name], thing=item)
  elif isinstance(thing, (list, tuple)):
    for index, item in enumerate(thing):
      if isinstance(item, BaseModel):
        yield from _WarnModelExtras(path=path + [index], thing=item)
  elif isinstance(thing, (list, tuple)):
    for index, item in enumerate(thing):
      if isinstance(item, BaseModel):
        yield from _WarnModelExtras(path=path + [index], thing=item)
  elif isinstance(thing, dict):
    for key, item in thing.items():
      if isinstance(item, BaseModel):
        yield from _WarnModelExtras(path=path + [key], thing=item)
  elif isinstance(thing, BaseModel):
    if thing.model_extra is not None:

      for key, value in thing.model_extra.items():
        yield _ExtraFieldWarning(
            path=path + [key],
            thing=value,
            message=
            f'Warning: Unknown field: {key} in {thing.__class__.__name__} at {path}'
        )
        yield from _WarnModelExtras(path=path + [key], thing=value)

    for key, value in thing.model_fields.items():
      yield from _WarnModelExtras(path=path + [key], thing=value)
  else:
    pass


async def _GetNewPath(*, parent_path: Path) -> Path:
  now = datetime.datetime.now(datetime.timezone.utc)
  name = now.strftime('%Y-%m-%d_%H-%M-%S_%f')
  path = parent_path / name
  index = 1
  while await path.exists():
    path = parent_path / f'{name}_{index}'
    index += 1
  await path.mkdir(parents=True, exist_ok=True)
  return path


async def TryParseAsModel(
    *,
    content: Any,
    model_type: Type[_BaseModelT],
    errors_dump_directory: Optional[Path],
    strict: Literal['yes', 'no', 'warn'] = 'warn') -> _BaseModelT:

  async def _Internal(errors_dump_directory: Optional[Path]):
    error_dump_path: Optional[Path] = None

    try:
      try:
        return await to_thread(model_type.model_validate,
                               content,
                               strict=strict in ['yes', 'warn'])
      except pydantic_core.ValidationError as e:
        if strict == 'yes':
          # Go to the except below which just prints the error.
          raise
        # Try parsing it non-strictly, and then warn about the error if it
        # succeeds. If it errors, then we'll go to the except below which
        # prints the error.
        model = await to_thread(model_type.model_validate,
                                content,
                                strict=False)
        if error_dump_path is None and errors_dump_directory is not None:
          error_dump_path = await _GetNewPath(parent_path=errors_dump_directory)

        if error_dump_path is not None:
          error_line = await DumpBigErrorStr(
              exception=e,
              max_lines=MAX_DUMP_LINES,
              path=error_dump_path /
              f'{slugify(str(model_type))}-error_str.txt')

          model_dump_yaml = await DumpBigYaml(
              to_thread(model.model_dump,
                        mode='json',
                        by_alias=True,
                        round_trip=True),
              max_lines=MAX_DUMP_LINES,
              path=error_dump_path /
              f'{slugify(str(model_type))}-model_dump.yaml')
          input_content_yaml = await DumpBigYaml(
              content,
              max_lines=MAX_DUMP_LINES,
              path=error_dump_path /
              f'{slugify(str(model_type))}-input_content.yaml')
        else:
          error_line = str(e)
          model_dump_yaml = await DumpModelToYAML(model)
          input_content_yaml = await DumpYaml(content)

        msg = f'Warning: Error parsing {model_type} with strict=True: {error_line}'
        logger.error(
            f'{msg}:'
            f'\n\n{textwrap.indent(str(e), prefix="  ")}'
            f'\nParsed Model:\n{textwrap.indent(model_dump_yaml, prefix="  ")}'
            f'\nInput content\n{textwrap.indent(input_content_yaml, prefix="  ")}'
            f'\n{msg}')
        return model
    except pydantic_core.ValidationError as e:
      # It failed strictly or non-strictly. Print the error and raise.

      if error_dump_path is None and errors_dump_directory is not None:
        error_dump_path = await _GetNewPath(parent_path=errors_dump_directory)

      if error_dump_path is not None:
        error_line = await DumpBigErrorStr(
            exception=e,
            max_lines=MAX_DUMP_LINES,
            path=error_dump_path / f'{slugify(str(model_type))}-error_str.txt')
      else:
        error_line = str(e)
      msg_summary = f'Error parsing {model_type}: {error_line}'
      msg = f'{msg_summary}'
      if error_dump_path is not None:
        input_content_yaml = await DumpBigYaml(
            content,
            max_lines=MAX_DUMP_LINES,
            path=error_dump_path /
            f'{slugify(str(model_type))}-input_content.yaml')
        msg += '\nInput content\n' + textwrap.indent(input_content_yaml,
                                                     prefix='  ')
        errors_yaml = await DumpBigYaml(
            e.errors(),
            max_lines=MAX_DUMP_LINES,
            path=error_dump_path / f'{slugify(str(model_type))}-errors.yaml')
        msg += '\nError details\n' + textwrap.indent(errors_yaml, prefix='  ')
      else:
        msg += '\nInput content\n' + textwrap.indent(await DumpYaml(content),
                                                     prefix='  ')
        msg += '\nError details\n' + textwrap.indent(await DumpYaml(e.errors()),
                                                     prefix='  ')

      msg += f'\n{msg_summary}'
      raise Exception(msg) from e

  model = await _Internal(errors_dump_directory=errors_dump_directory)
  extra_fields_warnings = _WarnModelExtras(path=[], thing=model)
  if strict == 'warn':
    for warning in extra_fields_warnings:
      logger.warn(warning.message)
  elif strict == 'yes':
    for warning in extra_fields_warnings:
      raise Exception(warning.message)
  return model


class WatchVar:

  def __init__(self, **kwargs):
    self._kwargs = kwargs

  async def __aenter__(self):
    pass

  async def __aexit__(self, exc_type, exc, tb):
    if exc is not None:
      if isinstance(exc, (KeyboardInterrupt, SystemExit, asyncio.CancelledError,
                          BaseException)):
        return

      extra: Dict[str, Any] = {}
      if exc_type is not None:
        extra['exc_type'] = type(exc_type).__name__
      extra['kwargs'] = self._kwargs
      extra['tb'] = traceback.format_tb(tb)

      logger.error(
          f'{type(self).__name__}: Error occurred: {json.dumps(str(exc))} ({type(exc).__name__})'
          f'\nTraceback:\n{textwrap.indent("".join(traceback.format_tb(tb)), "  ")}'
          f'\nYou are watching these variables:\n{textwrap.indent(await DumpYaml(self._kwargs), "  ")}',
          extra=extra)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import asyncio
import datetime
from abc import ABC, abstractmethod
from typing import (Dict, List, Literal, NamedTuple, Optional, Sequence, Tuple,
                    Union, overload)

from anyio import Path
from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Annotated

from .comfy_schema import APINodeID, APIWorkflowTicket

JobID = Annotated[str, 'JobID']


class Progress(NamedTuple):
  value: int
  max_value: int


class ExceptionInfo(NamedTuple):
  type: str
  message: str
  traceback: str
  attributes: Dict[str, str]


class JobStatus(BaseModel):
  model_config = ConfigDict(frozen=True)

  scheduled: Optional[datetime.datetime] = Field(
      ..., description='Time the job was scheduled with Comfy Catapult.')
  comfy_scheduled: Optional[datetime.datetime] = Field(
      ..., description='Time the job was scheduled with ComfyUI.')
  pending: Optional[datetime.datetime] = Field(
      ...,
      description='If/when the job is seen in the ComfyUI /queue as pending.')
  running: Optional[datetime.datetime] = Field(
      ...,
      description='If/when the job is seen in the ComfyUI /queue as running.')
  success: Optional[datetime.datetime] = Field(
      ...,
      description=
      'If/when the job is seen in the ComfyUI /history as successful.')
  errored: Optional[datetime.datetime] = Field(
      ...,
      description='If/when the job has encountered an unrecoverable error.')
  cancelled: Optional[datetime.datetime] = Field(
      ..., description='If/when the job was cancelled.')
  errors: List[ExceptionInfo]

  system_stats_check: Optional[datetime.datetime] = Field(
      None,
      description=
      'Last time system stats were successfully checked (while this job is not done).'
  )
  queue_check: Optional[datetime.datetime] = Field(
      None,
      description=
      'Last time /queue/job_id was successfully checked for (while this job is not done).'
  )

  ticket: Optional[APIWorkflowTicket] = Field(
      ..., description='The ticket returned from the ComfyUI API.')

  job_history: Optional[dict] = Field(
      None,
      description=
      'The history of the job. This is only set when the job is done and the history is successfully retrieved.'
  )

  def IsDone(self) -> bool:
    """Returns True if the job is no longer viable; either it finished or an error occurred etc.."""
    return (self.success is not None or self.errored is not None
            or self.cancelled is not None)

  def _replace(self, **kwargs):
    return self.model_copy(update=kwargs)


class ComfyCatapultBase(ABC):

  @abstractmethod
  def __init__(self):
    pass

  @abstractmethod
  async def __aenter__(self):
    raise NotImplementedError()

  @abstractmethod
  async def __aexit__(self, exc_type, exc, tb):
    raise NotImplementedError()

  @abstractmethod
  async def Close(self):
    raise NotImplementedError()

  @overload
  @abstractmethod
  async def Catapult(
      self,
      *,
      job_id: JobID,
      prepared_workflow: dict,
      important: Sequence[APINodeID],
      use_future_api: Literal[True],
      job_debug_path: Optional[Path] = None
  ) -> Tuple[JobStatus, 'asyncio.Future[dict]']:
    ...

  @overload
  @abstractmethod
  async def Catapult(self,
                     *,
                     job_id: JobID,
                     prepared_workflow: dict,
                     important: Sequence[APINodeID],
                     use_future_api: Literal[False] = False,
                     job_debug_path: Optional[Path] = None) -> dict:
    ...

  @abstractmethod
  async def Catapult(
      self,
      *,
      job_id: JobID,
      prepared_workflow: dict,
      important: Sequence[APINodeID],
      use_future_api: bool = False,
      job_debug_path: Optional[Path] = None
  ) -> Union[dict, Tuple[JobStatus, 'asyncio.Future[dict]']]:
    """Schedule a ComfyUI workflow job.

    Args:
        job_id (str): A unique identifier for the job. Note: This string must
          be unique, and must be slugified! Use python-slugify to slugify the
          string. Note: This is not the same string as the prompt_id returned
          from ComfyUI API. You can find that in the JobStatus.ticket returned
          from GetStatus().
        prepared_workflow (dict): Workflow to submit.
        important (List[APINodeID]): List of important nodes (e.g output nodes
          we are interested in).
        use_future_api (bool): Use the future API; returns a future that will
          resolve to the job history when the job is done.
        job_debug_path (Path, optional): Path to save debug information. If
          None, will use sensible defaults.

    Raises:
        WorkflowSubmissionError: Failed to submit workflow.

    Returns:
        Union[dict, Tuple[JobStatus, asyncio.Future[dict]]]: If use_future_api
        is True, returns a tuple of the job status and a future that will
        resolve to the job history when the job is done. Otherwise, returns the
        job history.
    """
    raise NotImplementedError()

  @abstractmethod
  async def GetStatus(
      self,
      *,
      job_id: JobID,
      poll: bool = False) -> 'Tuple[JobStatus, asyncio.Future[dict]]':
    """Get the status of a job.

    Args:
        job_id (str): The job id.
        poll (bool): If True, will poll ComfyUI for the latest status update.

    Returns:
        Tuple[JobStatus, asyncio.Future[dict]]: The status of the job, and a
          future that will resolve to the job history when the job is done.
    """

  @abstractmethod
  async def GetExceptions(self, *, job_id: JobID) -> List[Exception]:
    """List of exceptions that occurred during the job.

    Args:
        job_id (str): The job id.

    Returns:
        List[Exception]: List of exceptions that occurred during the job.
    """
    raise NotImplementedError()

  @abstractmethod
  async def CancelJob(self, *, job_id: JobID):
    """Cancel a job. No-op if the job is done. Will also try to cancel the job remotely.

    Warning: Currently, ComfyUI will interrupt the currently running job, even
    if it is not the job you are trying to cancel. This is a limitation of the
    ComfyUI API.

    Args:
        job_id (str): The job id.
    """
    raise NotImplementedError()

  @abstractmethod
  async def Resume(self,
                   *,
                   job_id: JobID,
                   prepared_workflow: dict,
                   important: Sequence[APINodeID],
                   status: JobStatus,
                   poll: bool,
                   errors: List[Exception] = [],
                   job_debug_path: Optional[Path] = None):
    """Resume a job. This is useful if another instance of the ComfyCatapultBase
    is created and the job is "forgotten".
    """
    raise NotImplementedError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import json
from typing import List, Tuple

import aioshutil
from anyio import Path

from ._internal.url_utils import ToParseResult, ValidateIsBasedURL
from .comfy_schema import ComfyUIPathTriplet
from .remote_file_api_base import RemoteFileAPIBase


async def _LocalFileURLToLocalPath(url: str) -> Path:
  url_pr = ToParseResult(url)
  if url_pr.scheme != 'file':
    raise ValueError(f'URL {json.dumps(url)} is not a file:// URL')
  if url_pr.netloc != '':
    raise ValueError(f'URL {json.dumps(url)} is not a local file URL')
  if url_pr.params != '':
    raise ValueError(f'URL {json.dumps(url)} has params')
  if url_pr.query != '':
    raise ValueError(f'URL {json.dumps(url)} has query')
  if url_pr.fragment != '':
    raise ValueError(f'URL {json.dumps(url)} has fragment')
  if not url_pr.path.startswith('/'):
    raise ValueError(f'URL {json.dumps(url)} has relative path')
  path = Path(url_pr.path)
  path = await path.resolve()
  if not path.is_absolute():
    raise ValueError(f'URL {json.dumps(url)} has relative path')
  return path


async def _IsLocalPathWeaklyRelative(*, base: Path, path: Path) -> bool:
  base = await base.resolve()
  path = await path.resolve()
  return path.is_relative_to(base)


async def _ValidateLocalPath(*, path: Path, any_bases: List[Path]) -> Path:
  for base in any_bases:
    if await _IsLocalPathWeaklyRelative(base=base, path=path):
      return path
  raise ValueError(
      f'Path {json.dumps(str(path))} is not relative to any of {any_bases}')


class LocalRemoteFileAPI(RemoteFileAPIBase):
  """This one uses file:/// protocol on the local system.

  It is probably faster. In the future, I hope to add other protocols, so this
  can be used with other a choice remote storage systems as transparently as
  possible.
  """

  def __init__(self, *, upload_to_bases: List[str],
               download_from_bases: List[str]):
    super().__init__()
    self._upload_to_bases = upload_to_bases
    self._download_from_bases = download_from_bases

  async def UploadFile(self, *, src_path: Path, untrusted_dst_url: str) -> str:
    trusted_dst_url: str = ValidateIsBasedURL(url=untrusted_dst_url,
                                              any_bases=self._upload_to_bases)
    scheme = ToParseResult(trusted_dst_url).scheme
    if scheme != 'file':
      raise ValueError(
          f'URL {json.dumps(trusted_dst_url)} is not a file:// URL')
    if not await src_path.exists():
      raise ValueError(f'File {src_path} does not exist')
    if not await src_path.is_file():
      raise ValueError(f'File {src_path} is not a file')
    untrusted_dst_path = await _LocalFileURLToLocalPath(url=trusted_dst_url)
    trusted_dst_path = await _ValidateLocalPath(
        path=untrusted_dst_path,
        any_bases=[
            await _LocalFileURLToLocalPath(url=base)
            for base in self._upload_to_bases
        ])

    await trusted_dst_path.parent.mkdir(parents=True, exist_ok=True)
    await aioshutil.copy(src_path, trusted_dst_path)
    return trusted_dst_url

  async def DownloadFile(self, *, untrusted_src_url: str, dst_path: Path):
    trusted_src_url: str = ValidateIsBasedURL(
        url=untrusted_src_url, any_bases=self._download_from_bases)
    scheme = ToParseResult(trusted_src_url).scheme
    if scheme != 'file':
      raise ValueError(
          f'URL {json.dumps(trusted_src_url)} is not a file:// URL')

    untrusted_src_path = await _LocalFileURLToLocalPath(url=trusted_src_url)
    trusted_src_path = await _ValidateLocalPath(
        path=untrusted_src_path,
        any_bases=[
            await _LocalFileURLToLocalPath(url=base)
            for base in self._download_from_bases
        ])
    if not await trusted_src_path.exists():
      raise ValueError(f'File {trusted_src_path} does not exist')
    if not await trusted_src_path.is_file():
      raise ValueError(f'File {trusted_src_path} is not a file')

    await dst_path.parent.mkdir(parents=True, exist_ok=True)
    await aioshutil.copy(trusted_src_path, dst_path)

  async def DownloadTriplet(self, *, untrusted_comfy_api_url: str,
                            untrusted_src_triplet: ComfyUIPathTriplet,
                            dst_path: Path):
    # TODO: Maybe we can make a mapper that maps triplets to urls, and then
    # download the file from the URL. But this is not necessary right now,
    # because you can use the ComfyAPIRemoteFileAPI to download triplets,
    # and the GenericRemoteFileAPI to handle it transparently by choosing
    # the RemoteAPIFileBase.
    raise NotImplementedError('Local files do not support triplets')

  async def UploadToTriplet(
      self, *, src_path: Path, untrusted_comfy_api_url: str,
      untrusted_dst_triplet: ComfyUIPathTriplet) -> ComfyUIPathTriplet:
    # TODO: Maybe we can make a mapper that maps triplets to urls, and then
    # download the file from the URL. But this is not necessary right now,
    # because you can use the ComfyAPIRemoteFileAPI to download triplets,
    # and the GenericRemoteFileAPI to handle it transparently by choosing
    # the RemoteAPIFileBase.
    raise NotImplementedError('Local files do not support triplets')

  def TripletToURL(self, *, comfy_api_url: str,
                   triplet: ComfyUIPathTriplet) -> str:
    raise NotImplementedError('Local files do not support triplets')

  def URLToTriplet(self, *, url: str) -> Tuple[str, ComfyUIPathTriplet]:
    raise NotImplementedError('Local files do not support triplets')

  def GetBases(self) -> List[str]:
    return list(self._upload_to_bases + self._download_from_bases)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import unittest

from ._internal.url_utils import JoinToBaseURL


class TestUrlUtils(unittest.TestCase):

  def test_JoinToBaseURL(self):
    self.assertEqual(JoinToBaseURL('http://example.com', 'path'),
                     'http://example.com/path')
    self.assertEqual(JoinToBaseURL('http://example.com/', 'path'),
                     'http://example.com/path')
    self.assertEqual(JoinToBaseURL('http://example.com', '/path'),
                     'http://example.com/path')
    self.assertEqual(JoinToBaseURL('http://example.com/', '/path'),
                     'http://example.com/path')

    self.assertEqual(JoinToBaseURL('http://example.com/path', 'to'),
                     'http://example.com/path/to')
    self.assertEqual(JoinToBaseURL('http://example.com/path/', 'to'),
                     'http://example.com/path/to')
    self.assertEqual(JoinToBaseURL('http://example.com/path', '/to'),
                     'http://example.com/path/to')
    self.assertEqual(JoinToBaseURL('http://example.com/path/', '/to'),
                     'http://example.com/path/to')


if __name__ == '__main__':
  unittest.main(buffer=True)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import json
from collections import defaultdict
from typing import Dict, List, Tuple

from anyio import Path

from ._internal.url_utils import IsWeaklyRelativeTo
from .comfy_schema import ComfyUIPathTriplet
from .remote_file_api_base import RemoteFileAPIBase


class GenericRemoteFileAPI(RemoteFileAPIBase):
  """Download or upload files from a URL, using multiple schemss.
  """

  def __init__(self):
    super().__init__()
    self._base_to_apis: Dict[str, List[RemoteFileAPIBase]] = defaultdict(list)

  def Register(self, api: RemoteFileAPIBase):
    for base in api.GetBases():
      self._base_to_apis[base].append(api)

  def _GetAPIsForURL(self, *, url: str) -> List[RemoteFileAPIBase]:
    apis: List[RemoteFileAPIBase] = []
    for base, apis in self._base_to_apis.items():
      if IsWeaklyRelativeTo(base=base, url=url):
        apis.extend(apis)
    if apis:
      return apis
    raise ValueError(f'URL {url} is not relative to any of'
                     f' {self._base_to_apis.keys()}, there is no API registered'
                     f' to handle such URLs')

  def _GetAPIsForTriplet(
      self, *, comfy_api_url: str,
      triplet: ComfyUIPathTriplet) -> List[RemoteFileAPIBase]:
    relevant_apis = []
    convered_urls = []
    for base, base_apis in self._base_to_apis.items():
      for api in base_apis:
        try:
          url = api.TripletToURL(comfy_api_url=comfy_api_url, triplet=triplet)
        except NotImplementedError:
          continue

        convered_urls.append({'base': base, 'url': url})
        if IsWeaklyRelativeTo(base=base, url=url):
          relevant_apis.append(api)
    if relevant_apis:
      return relevant_apis
    raise ValueError(f'ComfyUI API server URL {json.dumps(comfy_api_url)}:'
                     ' there is no API registered to handle this URL'
                     f'\n triplet: {triplet}'
                     f'\n convered_urls: {convered_urls}')

  async def UploadFile(self, *, src_path: Path, untrusted_dst_url: str) -> str:
    if not await src_path.exists():
      raise ValueError(f'File {src_path} does not exist')
    if not await src_path.is_file():
      raise ValueError(f'File {src_path} is not a file')

    apis: List[RemoteFileAPIBase] = self._GetAPIsForURL(url=untrusted_dst_url)
    for i, api in enumerate(apis):
      try:
        return await api.UploadFile(src_path=src_path,
                                    untrusted_dst_url=untrusted_dst_url)
      except ValueError:
        if i + 1 >= len(apis):
          raise
    raise AssertionError('unreachable')

  async def UploadToTriplet(
      self, *, src_path: Path, untrusted_comfy_api_url: str,
      untrusted_dst_triplet: ComfyUIPathTriplet) -> ComfyUIPathTriplet:
    if not await src_path.exists():
      raise ValueError(f'File {src_path} does not exist')
    if not await src_path.is_file():
      raise ValueError(f'File {src_path} is not a file')

    apis: List[RemoteFileAPIBase] = self._GetAPIsForTriplet(
        comfy_api_url=untrusted_comfy_api_url, triplet=untrusted_dst_triplet)

    for i, api in enumerate(apis):
      try:
        return await api.UploadToTriplet(
            untrusted_comfy_api_url=untrusted_comfy_api_url,
            src_path=src_path,
            untrusted_dst_triplet=untrusted_dst_triplet)
      except ValueError:
        if i + 1 >= len(apis):
          raise
    raise AssertionError('unreachable')

  async def DownloadFile(self, *, untrusted_src_url: str, dst_path: Path):
    apis: List[RemoteFileAPIBase] = self._GetAPIsForURL(url=untrusted_src_url)
    for i, api in enumerate(apis):
      try:
        return await api.DownloadFile(untrusted_src_url=untrusted_src_url,
                                      dst_path=dst_path)
      except ValueError:
        if i + 1 >= len(apis):
          raise
    raise AssertionError('unreachable')

  async def DownloadTriplet(self, *, untrusted_comfy_api_url: str,
                            untrusted_src_triplet: ComfyUIPathTriplet,
                            dst_path: Path):
    apis: List[RemoteFileAPIBase] = self._GetAPIsForTriplet(
        comfy_api_url=untrusted_comfy_api_url, triplet=untrusted_src_triplet)
    for i, api in enumerate(apis):
      try:
        return await api.DownloadTriplet(
            untrusted_comfy_api_url=untrusted_comfy_api_url,
            untrusted_src_triplet=untrusted_src_triplet,
            dst_path=dst_path)
      except ValueError:
        if i + 1 >= len(apis):
          raise
    raise AssertionError('unreachable')

  def TripletToURL(self, *, comfy_api_url: str,
                   triplet: ComfyUIPathTriplet) -> str:
    apis: List[RemoteFileAPIBase] = self._GetAPIsForTriplet(
        comfy_api_url=comfy_api_url, triplet=triplet)
    for i, api in enumerate(apis):
      try:
        return api.TripletToURL(comfy_api_url=comfy_api_url, triplet=triplet)
      except ValueError:
        if i + 1 >= len(apis):
          raise
    raise AssertionError('unreachable')

  def URLToTriplet(self, *, url: str) -> Tuple[str, ComfyUIPathTriplet]:
    apis: List[RemoteFileAPIBase] = self._GetAPIsForURL(url=url)
    for i, api in enumerate(apis):
      try:
        return api.URLToTriplet(url=url)
      except (NotImplementedError, ValueError):
        if i + 1 >= len(apis):
          raise
    raise AssertionError('unreachable')

  def GetBases(self) -> List[str]:
    return list(self._base_to_apis.keys())



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.

import json
import os
import unittest
import uuid
from typing import Dict, List, Optional
from unittest.mock import AsyncMock
from urllib.parse import urlparse

import aiofiles
from anyio import Path
from websockets import connect

from ._internal.utilities import (BasicAuthToHeaders, GetWebSocketURL,
                                  TryParseAsModel)
from .api_client import ComfyAPIClient
from .catapult import ComfyCatapult
from .catapult_base import JobStatus
from .comfy_schema import APIWorkflow, WSMessage
from .comfy_utils import GetNodeByTitle

COMFY_API_URL = os.environ.get('COMFY_API_URL')
if COMFY_API_URL is None:
  raise ValueError('Please set COMFY_API_URL in the environment')


class StrictMock(AsyncMock):

  def __getattr__(self, name):
    if name in self.__dict__:
      return super().__getattr__(name)
    raise AttributeError(f"Mock object has no attribute '{name}'")


class CatapultTest(unittest.IsolatedAsyncioTestCase):

  async def asyncSetUp(self):
    if COMFY_API_URL is None:
      raise ValueError('Please set COMFY_API_URL in the environment')
    self._comfy_api_url: str = COMFY_API_URL
    self._job_debug_path: Optional[Path] = None
    job_debug_path: Optional[str] = os.environ.get('JOB_DEBUG_PATH')
    if job_debug_path is not None:
      self._job_debug_path = Path(job_debug_path)

  async def test_WSBasicAuth(self):
    """
    This test is assumes that COMFY_API_URL is a basic auth URL. Otherwise, it
    does not actually test anything useful.
    """
    errors_dump_directory: Optional[Path] = None
    if self._job_debug_path is not None:
      errors_dump_directory = self._job_debug_path / 'errors'

    client_id = str(uuid.uuid4())
    ws_url = urlparse(
        GetWebSocketURL(comfy_api_url=self._comfy_api_url, client_id=client_id))

    ws_headers: Dict[str, str] = {}
    ws_url = urlparse(
        BasicAuthToHeaders(url=ws_url.geturl(), headers=ws_headers))

    print(f'ws_url: {ws_url.geturl()}')
    async with connect(ws_url.geturl(), extra_headers=ws_headers) as ws:
      while True:
        out = await ws.recv()
        if not isinstance(out, str):
          continue
        message = await TryParseAsModel(
            content=json.loads(out),
            model_type=WSMessage,
            errors_dump_directory=errors_dump_directory)
        print(message)
        return

  async def test_Resume(self):
    async with aiofiles.open('test_data/default_workflow_api.json', 'r') as f:
      prepared_workflow_dict = json.loads(await f.read())

    prepared_workflow = APIWorkflow.model_validate(prepared_workflow_dict)
    _, ksampler_node = GetNodeByTitle(workflow=prepared_workflow,
                                      title='KSampler')
    # Make this 100 so we have time to resume the job.
    ksampler_node.inputs['steps'] = 100

    job_id = str(uuid.uuid4())
    errors: List[Exception] = []

    async with ComfyAPIClient(
        comfy_api_url=self._comfy_api_url) as comfy_client:
      async with ComfyCatapult(comfy_client=comfy_client,
                               debug_path=None) as catapult:
        job_status: JobStatus
        job_status, _ = await catapult.Catapult(
            job_id=job_id,
            prepared_workflow=prepared_workflow_dict,
            important=[],
            use_future_api=True)

        errors = await catapult.GetExceptions(job_id=job_id)

    async with ComfyAPIClient(
        comfy_api_url=self._comfy_api_url) as comfy_client:
      async with ComfyCatapult(comfy_client=comfy_client,
                               debug_path=None) as catapult:
        await catapult.Resume(job_id=job_id,
                              prepared_workflow=prepared_workflow_dict,
                              important=[],
                              status=job_status,
                              poll=True,
                              errors=errors)

        job_status, future = await catapult.GetStatus(job_id=job_id)
        await future


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
#
# The Comfy Catapult project requires contributions made to this file be licensed
# under the MIT license or a compatible open source license. See LICENSE.md for
# the license text.
"""
CLI to run a ComfyUI API Workflow against a ComfyUI instance.
"""
import argparse
import asyncio
import json
import logging
import os
import sys
import uuid
import warnings
from datetime import datetime, timezone
from pathlib import Path
from shutil import get_terminal_size
from typing import List, Optional, Tuple

import anyio
import colorama
from attr import dataclass
from rich.console import Console
from rich_argparse import RichHelpFormatter
from slugify import slugify

from . import _build_version
from ._internal.utilities import DumpModelToDict, DumpModelToYAML
from .api_client import ComfyAPIClient
from .api_client_base import ComfyAPIClientBase
from .catapult import ComfyCatapult
from .catapult_base import ComfyCatapultBase, JobStatus
from .comfy_schema import APINodeID, APISystemStats
from .remote_file_api_comfy import ComfySchemeRemoteFileAPI
from .remote_file_api_generic import GenericRemoteFileAPI

logger = logging.getLogger(__name__)


def _GetProgramName() -> str:
  if __package__:
    # Use __package__ to get the base package name
    base_module_path = __package__
    # Infer the module name from the file path, with assumptions about the structure
    module_name = Path(__file__).stem
    # Construct what might be the intended full module path
    full_module_path = f'{base_module_path}.{module_name}' if base_module_path else module_name
    return f'python -m {full_module_path}'
  else:
    return sys.argv[0]


class _CustomRichHelpFormatter(RichHelpFormatter):

  def __init__(self, *args, **kwargs):
    if kwargs.get('width') is None:
      width, _ = get_terminal_size()
      if width == 0:
        warnings.warn('Terminal width was set to 0, using default width of 80.',
                      RuntimeWarning,
                      stacklevel=0)
        # This is the default in get_terminal_size().
        width = 80
      # This is what HelpFormatter does to the width returned by
      # `get_terminal_size()`.
      width -= 2
      kwargs['width'] = width
    super().__init__(*args, **kwargs)


async def GetWorkflow(workflow_path: str) -> str:
  if workflow_path == '-':
    workflow_json = sys.stdin.read()
  else:
    workflow_json = Path(workflow_path).read_text()
  return workflow_json


async def GetRemote(*, comfy_api_url: str):
  # Utility to help download/upload files.
  remote = GenericRemoteFileAPI()
  # This maps comfy+http://comfy_host:port/folder_type/subfolder/filename to
  # the /view and /upload API endpoints.
  remote.Register(
      ComfySchemeRemoteFileAPI(comfy_api_urls=[comfy_api_url], overwrite=True))
  # if args.comfy_install_file_url is not None:
  #   scheme = ToParseResult(args.comfy_install_file_url).scheme
  #   if scheme != 'file':
  #     raise ValueError(
  #         f'args.comfy_install_file_url must be a file:// URL, but is {args.comfy_install_file_url}'
  #     )

  #   # This one uses file:/// protocol on the local system. It is probably
  #   # faster. In the future, I hope to add other protocols, so this can be
  #   # used with other a choice remote storage systems as transparently as
  #   # possible.
  #   remote.Register(
  #       LocalRemoteFileAPI(upload_to_bases=[args.comfy_input_file_url],
  #                          download_from_bases=[
  #                              args.comfy_output_file_url,
  #                              args.comfy_temp_file_url
  #                          ]))
  return remote


async def DumpInfo(comfy_client: ComfyAPIClientBase,
                   catapult: ComfyCatapultBase, job_id: str, console: Console):

  status: JobStatus
  status, _ = await catapult.GetStatus(job_id=job_id)
  console.print(await DumpModelToDict(status))

  system_stats: APISystemStats = await comfy_client.GetSystemStats()
  console.print('system_stats:', style='bold blue')
  console.print(await DumpModelToYAML(system_stats))


async def StatusThread(stop_event: asyncio.Event,
                       comfy_client: ComfyAPIClientBase,
                       catapult: ComfyCatapultBase, job_id: str,
                       console: Console):
  while not stop_event.is_set():
    try:
      await asyncio.sleep(5)
      await DumpInfo(comfy_client=comfy_client,
                     catapult=catapult,
                     job_id=job_id,
                     console=console)
    except Exception as e:
      console.print(f'Error in StatusThread: {e}', style='bold red')
      console.print_exception()


def ParseArgs() -> Tuple[argparse.ArgumentParser, argparse.Namespace]:
  p = argparse.ArgumentParser(prog=_GetProgramName(),
                              description=__doc__,
                              formatter_class=_CustomRichHelpFormatter)
  p.add_argument('--version', action='version', version=_build_version)
  p.add_argument(
      '--comfy-api-url',
      type=str,
      default=None,
      help=
      'URL of the ComfyUI API. If left empty, will default to COMFY_API_URL.')
  p.add_argument('--debug-path',
                 type=Path,
                 default=Path('.debug'),
                 help='Path to save debug information.')

  # Add commands: execute, execute-bg, status, wait, cancel.

  sub_p = p.add_subparsers(dest='command',
                           title='commands',
                           description='Choose a command to run.',
                           required=True)

  p_execute = sub_p.add_parser('execute',
                               help='Execute a workflow.',
                               formatter_class=_CustomRichHelpFormatter)

  p_execute.add_argument('--job-id',
                         type=str,
                         default=None,
                         help='Unique job identifier.')
  p_execute.add_argument('--workflow-path',
                         type=str,
                         required=True,
                         help='Input markdown file, use "-" for stdin.')
  args = p.parse_args()
  return p, args


async def amain():
  console = Console(file=sys.stderr)
  args: Optional[argparse.Namespace] = None
  try:
    # Windows<10 requires this.
    colorama.init()

    p, args = ParseArgs()

    job_id: Optional[str] = args.job_id
    workflow_path: str = args.workflow_path
    comfy_api_url: Optional[str] = args.comfy_api_url
    debug_path: anyio.Path = anyio.Path(args.debug_path)

    if comfy_api_url is None:
      comfy_api_url = os.environ.get('COMFY_API_URL')
    if comfy_api_url is None or comfy_api_url == '':
      p.error(
          'comfy-api-url is required, or set the COMFY_API_URL environment variable.'
      )
      return

    workflow_template_json_str: str = await GetWorkflow(
        workflow_path=workflow_path)
    workflow_template_dict: dict = json.loads(workflow_template_json_str)

    async with ComfyAPIClient(comfy_api_url=comfy_api_url) as comfy_client:

      remote = await GetRemote(comfy_api_url=comfy_api_url)

      # Dump the ComfyUI server stats.
      system_stats: APISystemStats = await comfy_client.GetSystemStats()
      console.print('system_stats:', style='bold blue')
      console.print(await DumpModelToYAML(system_stats))

      async with ComfyCatapult(comfy_client=comfy_client,
                               debug_path=debug_path / 'catapult',
                               debug_save_all=True) as catapult:

        dt_str = datetime.now(tz=timezone.utc).isoformat()

        if job_id is None:
          job_id = f'{slugify(dt_str)}-my-job-{uuid.uuid4()}'

        job_info = GenericWorkflowInfo(
            client=comfy_client,
            catapult=catapult,
            remote=remote,
            workflow_template_dict=workflow_template_dict,
            workflow_dict=json.loads(workflow_template_json_str),
            important=[],
            job_id=job_id,
            job_history_dict=None,
            comfy_api_url=comfy_api_url)

        stop_event = asyncio.Event()
        status_thread_future = asyncio.create_task(
            StatusThread(stop_event=stop_event,
                         comfy_client=comfy_client,
                         catapult=catapult,
                         job_id=job_id,
                         console=console))

        await RunWorkflow(job_info=job_info)
        stop_event.set()
        await status_thread_future

        console.print('Job complete, now dumping info', style='bold green')

        await DumpInfo(comfy_client=comfy_client,
                       catapult=catapult,
                       job_id=job_id,
                       console=console)

        console.print('All done', style='bold green')

  except Exception:
    console.print_exception()
    if args:
      console.print('args:', args._get_kwargs(), style='bold red')

    sys.exit(1)
    return


@dataclass
class GenericWorkflowInfo:
  client: ComfyAPIClient
  catapult: ComfyCatapult
  remote: GenericRemoteFileAPI
  workflow_template_dict: dict
  workflow_dict: dict
  important: list
  job_id: str
  job_history_dict: Optional[dict]
  comfy_api_url: str


async def PrepareWorkflow(*, job_info: GenericWorkflowInfo):
  # You have to write this function, to change the workflow_dict as you like.
  pass


async def DownloadResults(*, job_info: GenericWorkflowInfo):
  # You have to write this function, to download the results you care about.
  pass


async def RunWorkflow(*, job_info: GenericWorkflowInfo):

  # You have to write this function, to change the workflow_dict as you like.
  await PrepareWorkflow(job_info=job_info)

  job_id: str = job_info.job_id
  workflow_dict: dict = job_info.workflow_dict
  important: List[APINodeID] = job_info.important

  # Here the magic happens, the job is submitted to the ComfyUI server.
  job_info.job_history_dict = await job_info.catapult.Catapult(
      job_id=job_id, prepared_workflow=workflow_dict, important=important)

  # Now that the job is done, you have to write something that will go and get
  # the results you care about, if necessary.
  await DownloadResults(job_info=job_info)


def main():
  asyncio.run(amain())


if __name__ == '__main__':
  main()
